"""
 PATTERN:
 dependency injection
"""
import importlib.util


class ModuleLazyLoader:

    def __init__(self, mod_name: str, pypath: str):
        self._m = None
        self._name = mod_name
        self._pypath = pypath

    @property
    def ready(self):
        return self._m is not None

    @property
    def pymod(self):  # ensure mod is ready before calling this
        return self._m

    @property
    def name(self):
        return self._name

    def load_now(self, parg):
        nam = self.name
        if nam != 'pygame':
            print(f' lazy loading: {nam}')  # debug infos

        if self._pypath[0] == '.':  # relative import detected
            self._m = importlib.import_module(self._pypath, parg)
        else:
            self._m = importlib.import_module(self._pypath)


class Injector:
    """
    gere chargement composants engine
    + any engine plugin
    Le fait au moyen dun tableau:
    module name <-> instance of LazyModule
    """

    def __init__(self, module_listing, pack_arg='katagames_engine'):
        self._listing = dict()
        for mname, pypath in module_listing.items():
            obj = ModuleLazyLoader(mname, pypath)
            self._listing[obj.name] = obj
        self._man_set = dict()

        self.package_arg = pack_arg
        self._loading_done = False

    def __contains__(self, item):
        if item in self._man_set:
            return True
        else:
            return item in self._listing

    def set(self, mname, pymod):
        self._man_set[mname] = pymod

    def __getitem__(self, item):
        if item in self._man_set:
            return self._man_set[item]
        else:
            tmp = self._listing[item]
            if not tmp.ready:
                self._loading_done = True
                tmp.load_now(self.package_arg)
            return tmp.pymod

    def register(self, sm_name, py_path):
        if self._loading_done:
            print('***warning*** register plugin should be done before using loading elements')
        self._listing[sm_name] = ModuleLazyLoader(sm_name, py_path)





class DijkstraPathfinder:
    @staticmethod
    def find_path(espace_matr, source_pos, dest_pos, limitations_wh=None):
        """
        :param espace_matr:
        :param source_pos:
        :param dest_pos:
        :param limitations_wh:
        :return: soit None, soit une liste de coords
        """
        chemin_inv = dict()  # on anticipe le backtrace code, en vue de retourner le plus court chemin
        candidats = set()
        src = tuple(source_pos)
        dest = tuple(dest_pos)

        # --- initialisation distance à tous les pts
        dist = dict()
        if limitations_wh:
            limit_x, limit_y = limitations_wh
        else:
            limit_x, limit_y = espace_matr.get_size()

        binf_x, bsup_x = src[0] - limit_x, src[0] + limit_x
        binf_y, bsup_y = src[1] - limit_y, src[1] + limit_y
        for x in range(binf_x, bsup_x + 1):
            for y in range(binf_y, bsup_y + 1):
                k = (x, y)
                if espace_matr.is_out(*k):
                    continue
                dist[k] = 99999
                candidats.add(k)
        dist[src] = 0

        visited = set()  # représente l'ensemble des sommets déjà visités

        while len(candidats) > 0:
            # select the element of Q with the min. distance
            # cest à dire on va explorer le sommet listé dans Q, de distance minimale
            dist_minimale = 999999
            u = None
            for sommet_a_explorer in candidats:
                if dist[sommet_a_explorer] < dist_minimale:
                    dist_minimale = dist[sommet_a_explorer]
                    u = sommet_a_explorer

            # on indique que u est exploré
            candidats.remove(u)
            visited.add(u)

            # exploration : calcul des voisins de u
            voisins = list()
            les_possib = [
                (u[0] - 1, u[1]),
                (u[0] + 1, u[1]),
                (u[0], u[1] + 1),
                (u[0], u[1] - 1),
            ]
            for k in les_possib:
                if espace_matr.is_out(*k):
                    continue
                if not (binf_x <= k[0] <= bsup_x):
                    continue
                if not (binf_y <= k[1] <= bsup_y):
                    continue
                if espace_matr.get_val(*k):
                    continue  # blocage
                voisins.append(k)

            for v in voisins:  # m-à-j plus courtes distances pr arriver à v
                w = 1
                if dist[v] > dist[u] + w:
                    dist[v] = dist[u] + w
                    chemin_inv[v] = u

        # tous les candidats ont été testés => backtrace code c-à-d reconstruction chemin le + court
        chemin = [dest]
        x = dest
        no_way = False
        while x != src:
            if x not in chemin_inv:
                no_way = True
                break
            x = chemin_inv[x]
            chemin.insert(0, x)  # insertion en 1ère position

        if no_way:
            return None
        return chemin



from . import frects
from . import image
from ... import _hub
pygame = _hub.pygame

from .general import my_state, TEXT_COLOR, Border, default_border, wrap_with_records
from .image import wrap_multi_line, draw_text

# respond_event: Receives an event.
#   If the widget has a method corresponding to the event,
#   that method will be called.

# Note that the widget needs to be added to my_state.widgets to be used!
# Or, as the child of another widget. Removing it from that list
# removes it.

ACTIVE_FLASH = [(0, 0, 0), ] * 5 + [(230, 230, 0), ] * 20

widget_border_off = Border(border_width=8, tex_width=16, border_name="sys_widbor_edge1.png",
                           tex_name="sys_widbor_back.png", tl=0, tr=3, bl=4, br=5, t=1, b=1, l=2, r=2, padding=2)
widget_border_on = Border(border_width=8, tex_width=16, border_name="sys_widbor_edge2.png",
                          tex_name="sys_widbor_back.png", tl=0, tr=3, bl=4, br=5, t=1, b=1, l=2, r=2, padding=2)
popup_menu_border = Border(border_width=8, tex_width=16, border_name="sys_widbor_edge2.png",
                           tex_name="sys_defbackground.png", tl=0, tr=3, bl=4, br=5, t=1, b=1, l=2, r=2, padding=8,
                           transparent=False)


class Widget(frects.Frect):
    def __init__(self, dx, dy, w, h, data=None, on_click=None, tooltip=None, children=(), active=True,
                 show_when_inactive=False, on_right_click=None, **kwargs):
        # on_click takes widget, event as parameters.
        # on_right_click takes widget, event as parameters.
        super().__init__(dx, dy, w, h, **kwargs)
        self.data = data
        self.active = active
        self.tooltip = tooltip
        self.on_click = on_click
        self.on_right_click = on_right_click
        self.children = list(children)
        self.show_when_inactive = show_when_inactive

    def respond_event(self, ev):
        if self.active:
            for c in self.children:
                c.respond_event(ev)
            if self.get_rect().collidepoint(my_state.mouse_pos):
                if self.active and (ev.type == pygame.MOUSEBUTTONUP) and (
                        ev.button == 1) and not my_state.widget_clicked:
                    if not my_state.widget_clicked:
                        my_state.active_widget = self
                    if self.on_click:
                        self.on_click(self, ev)
                    my_state.widget_clicked = True
                elif self.active and (ev.type == pygame.MOUSEBUTTONUP) and (
                        ev.button == 3) and self.on_right_click and not my_state.widget_clicked:
                    if not my_state.widget_clicked:
                        my_state.active_widget = self
                    self.on_right_click(self, ev)
                    my_state.widget_clicked = True
            elif my_state.active_widget is self:
                if self.on_click and (ev.type == pygame.KEYDOWN) and (ev.key in my_state.get_keys_for("click_widget")):
                    self.on_click(self, ev)
                    my_state.widget_clicked = True
            self._builtin_responder(ev)

    def _builtin_responder(self, ev):
        pass

    def super_render(self):
        # This renders the widget and children, setting tooltip and whatnot.
        if self.active or self.show_when_inactive:
            self.render(self._should_flash())
            if self.tooltip and self.get_rect().collidepoint(my_state.mouse_pos):
                my_state.widget_tooltip = self.tooltip
            for c in self.children:
                c.super_render()

    def _should_flash(self):
        if self.parent and hasattr(self.parent, "kb_flash_override"):
            return self.parent.kb_flash_override(self)
        elif hasattr(self, "kbhandler") and self.kbhandler and hasattr(self.kbhandler, "kb_flash_override"):
            return self.kbhandler.kb_flash_override(self)
        else:
            return (self is my_state.active_widget) and my_state.active_widget_hilight

    def _default_flash(self):
        pygame.draw.rect(my_state.screen, ACTIVE_FLASH[my_state.anim_phase % len(ACTIVE_FLASH)], self.get_rect(), 1)

    def render(self, flash=False):
        if flash:
            self._default_flash()

    def is_kb_selectable(self):
        if self.parent and hasattr(self.parent, "kb_select_override"):
            return self.parent.kb_select_override(self)
        elif hasattr(self, "kbhandler") and self.kbhandler and hasattr(self.kbhandler, "kb_select_override"):
            return self.kbhandler.kb_select_override(self)
        else:
            return self.on_click


class ButtonWidget(Widget):
    def __init__(self, dx, dy, w, h, sprite=None, frame=0, on_frame=0, off_frame=0, **kwargs):
        super(ButtonWidget, self).__init__(dx, dy, w, h, **kwargs)
        self.sprite = sprite
        self.frame = frame
        self.on_frame = on_frame
        self.off_frame = off_frame

    def render(self, flash=False):
        if self.sprite:
            self.sprite.render(self.get_rect(), self.frame)
        if flash:
            self._default_flash()


class LabelWidget(Widget):
    def __init__(self, dx, dy, w, h, text='***', color=None, font=None, justify=-1, draw_border=False,
                 border=widget_border_off, text_fun=None, **kwargs):
        # text_fun is a function that takes this widget as a parameter. It returns the text to display.
        super(LabelWidget, self).__init__(dx, dy, w, h, **kwargs)
        self.text = text
        self.color = color or TEXT_COLOR
        self.font = font or my_state.small_font
        if h == 0:
            self.h = len(wrap_multi_line(text, self.font, self.w)) * self.font.get_linesize()
        self.justify = justify
        self.draw_border = draw_border
        self.border = border
        self.text_fun = text_fun

    def render(self, flash=False):
        if self.draw_border:
            self.border.render(self.get_rect())
        if self.text_fun:
            self.text = self.text_fun(self)
        draw_text(self.font, self.text, self.get_rect(), self.color, self.justify)
        if flash:
            self._default_flash()


class TextEntryWidget(Widget):
    ALLOWABLE_CHARACTERS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890()-=_+,.?"'

    def __init__(self, dx, dy, w, h, text='***', color=None, font=None, justify=-1, on_change=None, **kwargs):
        # on_change is a callable that takes (widget,ev) whenever the contents of the text changes.
        super(TextEntryWidget, self).__init__(dx, dy, w, h, **kwargs)
        if not text:
            text = ''
        self.char_list = list(text)
        self.color = color or TEXT_COLOR
        self.font = font or my_state.big_font
        self.justify = justify
        self.input_cursor = image.Image("sys_textcursor.png", 8, 16)
        self.on_change = on_change

    def render(self, flash=False):
        mydest = self.get_rect()
        if self is my_state.active_widget:
            widget_border_on.render(mydest.inflate(-4, -4))
        else:
            widget_border_off.render(mydest.inflate(-4, -4))
        myimage = self.font.render(self.text, True, self.color)
        my_state.screen.set_clip(mydest)
        textdest = myimage.get_rect(center=mydest.center)
        my_state.screen.blit(myimage, textdest)
        my_state.screen.set_clip(None)
        if flash or (my_state.active_widget is self):
            self.input_cursor.render(textdest.topright, (my_state.anim_phase // 3) % 4)

    def _builtin_responder(self, ev):
        if my_state.active_widget is self:
            if ev.type == pygame.KEYDOWN:
                if (ev.key == pygame.K_BACKSPACE) and (len(self.char_list) > 0):
                    del self.char_list[-1]
                    if self.on_change:
                        self.on_change(self, ev)
                elif (ev.unicode in self.ALLOWABLE_CHARACTERS) and (len(ev.unicode) > 0):
                    self.char_list.append(ev.unicode)
                    if self.on_change:
                        self.on_change(self, ev)

    def is_kb_selectable(self):
        return True

    def _get_text(self):
        return "".join(self.char_list)

    def _set_text(self, text):
        self.char_list = list(text)
        if self.on_change:
            self.on_change(self, None)

    text = property(_get_text, _set_text)

    def quietly_set_text(self, text):
        # Set the text without calling on_change.
        self.char_list = list(text)


class RadioButtonWidget(Widget):
    def __init__(self, dx, dy, w, h, sprite=None, buttons=(), spacing=2, **kwargs):
        # buttons is a list of dicts possibly containing: on_frame, off_frame, on_click, on_right_click, tooltip
        super().__init__(dx, dy, w, h, **kwargs)
        self.sprite = sprite
        self.buttons = list()
        self.spacing = spacing
        ddx = 0
        for b in buttons:
            self.buttons.append(ButtonWidget(
                ddx, 0, sprite.frame_width, sprite.frame_height, sprite,
                frame=b.get("off_frame", 1), on_frame=b.get("on_frame", 0),
                off_frame=b.get("off_frame", 1), tooltip=b.get("tooltip", None),
                on_click=self.click_radio, data=b.get("on_click", None),
                on_right_click=b.get("on_right_click", None),
                parent=self, anchor=frects.ANCHOR_UPPERLEFT
            ))
            ddx += sprite.frame_width + self.spacing
        self.buttons[0].frame = self.buttons[0].on_frame
        self.active_button = self.buttons[0]
        self.children += self.buttons

    def activate_button(self, button):
        self.active_button.frame = self.active_button.off_frame
        self.active_button = button
        button.frame = button.on_frame

    def get_button(self, data_sought):
        for b in self.buttons:
            if b.data == data_sought:
                return b

    def click_radio(self, button, ev):
        self.activate_button(button)
        if button.data:
            button.data(button, ev)


class ColumnWidget(Widget):
    def __init__(self, dx, dy, w, h, draw_border=False, border=default_border, padding=5, center_interior=False,
                 **kwargs):
        super(ColumnWidget, self).__init__(dx, dy, w, h, **kwargs)
        self.draw_border = draw_border
        self.border = border
        self._interior_widgets = list()
        self._header_widget = None
        self.padding = padding
        self.center_interior = center_interior

    def add_interior(self, other_w):
        self.children.append(other_w)
        self._interior_widgets.append(other_w)
        # Set the position of other_w inside this widget.
        other_w.parent = self
        self._position_contents()

    def set_header(self, other_w):
        self.children.append(other_w)
        self._header_widget = other_w
        # Set the position of other_w inside this widget.
        other_w.parent = self
        self._position_contents()

    def clear(self):
        del self._interior_widgets[:]
        del self.children[:]
        self._header_widget = None

    def _position_contents(self):
        dy = 0
        if self._header_widget:
            self._header_widget.dx = -(self._header_widget.w // 2)
            self._header_widget.dy = -(self._header_widget.h // 2)
            self._header_widget.anchor = frects.ANCHOR_TOP
            dy += self._header_widget.h // 2 + self.padding
        for widg in self._interior_widgets:
            if self.center_interior:
                widg.dx = (self.w - widg.w) // 2
            else:
                widg.dx = 0
            widg.dy = dy
            widg.anchor = frects.ANCHOR_UPPERLEFT
            dy += widg.h + self.padding
        self.h = dy

    def render(self, flash=False):
        if self.draw_border:
            self.border.render(self.get_rect())
        if flash:
            self._default_flash()


class ScrollColumnWidget(Widget):
    def __init__(self, dx, dy, w, h, up_button, down_button, draw_border=False, border=default_border, padding=5,
                 autoclick=False, focus_locked=False, **kwargs):
        super(ScrollColumnWidget, self).__init__(dx, dy, w, h, **kwargs)
        self.draw_border = draw_border
        self.border = border
        self._interior_widgets = list()
        self.padding = padding
        self.top_widget = 0
        self.up_button = up_button
        self.up_button.on_click = self.scroll_up
        self.up_button.kbhandler = self
        self.down_button = down_button
        self.down_button.on_click = self.scroll_down
        self.down_button.frame = self.down_button.off_frame
        self.down_button.kbhandler = self
        self.autoclick = autoclick
        self.focus_locked = focus_locked
        self._active_widget = -1
        self.active_widget = 0

    def _set_active_widget(self, widindex):
        if widindex >= 0 and widindex < len(self._interior_widgets) and widindex != self._active_widget:
            self._active_widget = widindex
            wid = self._interior_widgets[widindex]
            if not wid.active:
                self.scroll_to_index(widindex)
            if self.autoclick and wid.on_click:
                wid.on_click(wid, None)

    def _get_active_widget(self):
        return self._active_widget

    active_widget = property(_get_active_widget, _set_active_widget)

    def kb_select_override(self, child):
        return not (child in self._interior_widgets or child is self.up_button or child is self.down_button)

    def kb_flash_override(self, child):
        return ((my_state.active_widget is self) or self.focus_locked) and (child in self._interior_widgets) and (
                self._interior_widgets.index(child) == self.active_widget
        )

    def is_kb_selectable(self):
        return True

    def decorate_click(self, other_on_click):
        def nuclick(wid, ev):
            if wid in self._interior_widgets and self.active_widget != self._interior_widgets.index(wid):
                self.active_widget = self._interior_widgets.index(wid)
            other_on_click(wid, ev)

        return nuclick

    def add_interior(self, other_w):
        self.children.append(other_w)
        self._interior_widgets.append(other_w)
        # Set the position of other_w inside this widget.
        other_w.parent = self
        other_w.dx = 0
        other_w.anchor = frects.ANCHOR_UPPERLEFT
        self._position_contents()
        if other_w.on_click:
            other_w.on_click = self.decorate_click(other_w.on_click)

    def clear(self):
        for w in list(self._interior_widgets):
            self._interior_widgets.remove(w)
            self.children.remove(w)

    def _position_contents(self):
        # Disable all interior widgets, except those currently visible.
        for w in self._interior_widgets:
            w.active = False
        dy = 0
        n = self.top_widget
        if n >= len(self._interior_widgets):
            n = 0
        while (dy < self.h) and (n < len(self._interior_widgets)):
            widg = self._interior_widgets[n]
            widg.dy = dy
            if (dy == 0) or (dy + widg.h + self.padding <= self.h):
                widg.active = True
            dy += widg.h + self.padding
            n += 1
        # Activate or deactivate the up/down buttons.
        if self.top_widget > 0:
            self.up_button.frame = self.up_button.on_frame
        else:
            self.up_button.frame = self.up_button.off_frame
        if self._interior_widgets and not self._interior_widgets[-1].active:
            self.down_button.frame = self.down_button.on_frame
        else:
            self.down_button.frame = self.down_button.off_frame

        if self.active_widget < self.top_widget:
            self.active_widget = self.top_widget
        elif self.active_widget >= (n - 1):
            self.active_widget = max(n - 2, 0)

    def scroll_up(self, *args):
        if self.top_widget > 0:
            self.top_widget -= 1
            self._position_contents()

    def scroll_down(self, *args):
        if self._interior_widgets and not self._interior_widgets[-1].active:
            self.top_widget += 1
            self._position_contents()

    def scroll_to_index(self, index):
        '''Programmatic access to ensure a particular list item is shown'''
        if index < len(self._interior_widgets) and not self._interior_widgets[index].active:
            self.top_widget = index
            self._position_contents()
            self.active_widget = index

    def sort(self, key=None):
        self._interior_widgets.sort(key=key)
        self._position_contents()

    def _builtin_responder(self, ev):
        if (ev.type == pygame.MOUSEBUTTONDOWN) and self.get_rect().collidepoint(my_state.mouse_pos):
            if (ev.button == 4):
                self.scroll_up()
            elif (ev.button == 5):
                self.scroll_down()
        elif ((my_state.active_widget is self) or self.focus_locked) and (ev.type == pygame.KEYDOWN):
            if ev.key in my_state.get_keys_for("click_widget"):
                if self.active_widget < len(self._interior_widgets):
                    mybutton = self._interior_widgets[self.active_widget]
                    mybutton.on_click(mybutton, ev)
                    my_state.widget_clicked = True
            elif ev.key in my_state.get_keys_for("up") and self.active_widget > 0:
                self.active_widget -= 1
            elif ev.key in my_state.get_keys_for("down") and self.active_widget < (len(self._interior_widgets) - 1):
                self.active_widget += 1

    def render(self, flash=False):
        if self.draw_border:
            self.border.render(self.get_rect())
        if flash:
            self._default_flash()


class RowWidget(Widget):
    def __init__(self, dx, dy, w, h, draw_border=False, border=default_border, padding=5, **kwargs):
        super().__init__(dx, dy, w, h, **kwargs)
        self.draw_border = draw_border
        self.border = border
        self._left_widgets = list()
        self._center_widgets = list()
        self._right_widgets = list()
        self.padding = padding

    def add_left(self, other_w):
        self.children.append(other_w)
        self._left_widgets.append(other_w)
        # Set the position of other_w inside this widget.
        other_w.parent = self
        self._position_contents()

    def add_center(self, other_w):
        self.children.append(other_w)
        self._center_widgets.append(other_w)
        # Set the position of other_w inside this widget.
        other_w.parent = self
        self._position_contents()

    def add_right(self, other_w):
        self.children.append(other_w)
        self._right_widgets.append(other_w)
        # Set the position of other_w inside this widget.
        other_w.parent = self
        self._position_contents()

    def _position_contents(self):
        dx = 0
        self.h = max(w.h for w in self._right_widgets + self._center_widgets + self._left_widgets)
        for widg in self._left_widgets:
            widg.dx = dx
            widg.dy = -widg.h // 2
            widg.anchor = frects.ANCHOR_LEFT
            dx += widg.w + self.padding

        if self._center_widgets:
            dx = -(sum(w.w for w in self._center_widgets) + (len(self._center_widgets) - 1) * self.padding) // 2
            for widg in self._center_widgets:
                widg.dx = dx
                widg.dy = -widg.h // 2
                widg.anchor = frects.ANCHOR_CENTER
                dx += widg.w + self.padding

        if self._right_widgets:
            dx = -sum(w.w for w in self._right_widgets) - (len(self._right_widgets) - 1) * self.padding
            for widg in self._right_widgets:
                widg.dx = dx
                widg.dy = -widg.h // 2
                widg.anchor = frects.ANCHOR_RIGHT
                dx += widg.w + self.padding

    def render(self, flash=False):
        if self.draw_border:
            self.border.render(self.get_rect())
        if flash:
            self._default_flash()


class DropdownWidget(Widget):
    MENU_HEIGHT = 150

    def __init__(self, dx, dy, w, h, color=None, font=None, justify=-1, on_select=None, **kwargs):
        # on_select is a callable that takes the menu query result as its argument
        super(DropdownWidget, self).__init__(dx, dy, w, h, **kwargs)
        self.color = color or TEXT_COLOR
        self.font = font or my_state.small_font
        self.on_select = on_select
        self.on_click = self.open_menu
        self.menu = rpgmenu.Menu(dx, dy, w, self.MENU_HEIGHT, border=popup_menu_border, font=font,
                                 anchor=frects.ANCHOR_UPPERLEFT)

    def render(self, flash=False):
        mydest = self.get_rect()
        if self is my_state.active_widget:
            widget_border_on.render(mydest.inflate(-4, -4))
        else:
            widget_border_off.render(mydest.inflate(-4, -4))
        myimage = self.font.render(str(self.menu.get_current_item()), True, self.color)
        my_state.screen.set_clip(mydest)
        textdest = myimage.get_rect(center=mydest.center)
        my_state.screen.blit(myimage, textdest)
        my_state.screen.set_clip(None)
        if flash:
            self._default_flash()

    def add_item(self, msg, value, desc=None):
        self.menu.add_item(msg, value, desc)

    def open_menu(self, also_self_probably, ev):
        mydest = self.get_rect()
        mydest.h = self.menu.h
        mydest.w = self.menu.w
        mydest.clamp_ip(my_state.screen.get_rect())
        self.menu.dx, self.menu.dy = mydest.x, mydest.y
        result = self.menu.query()
        if self.on_select:
            self.on_select(result)

    @property
    def value(self):
        return self.menu.get_current_value()


class TextEditorWidget(Widget):
    # This widget is going to be different from the text entry widget above in that it's meant to be a full
    # functioned text editor. What have I gotten myself into?
    ALLOWABLE_CHARACTERS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890()-=_+,.?"\''

    def __init__(self, dx, dy, w, h, text='***', color=None, font=None, justify=-1, on_change=None, **kwargs):
        # on_change is a callable that takes (widget,ev) whenever the contents of the text changes.
        super().__init__(dx, dy, w, h, **kwargs)
        if not text:
            text = ''
        self.char_list = list(text)
        self.color = color or TEXT_COLOR
        self.font = font or my_state.big_font
        self.justify = justify
        self.cursor_image = image.Image("sys_editcursor.png", 8, 16)

        up_arrow = ButtonWidget(w - 32, -8, 32, 16, sprite=image.Image("sys_updownbuttons_small.png", 32, 16),
                                on_frame=0, off_frame=1, parent=self)
        down_arrow = ButtonWidget(w - 32, h + 8, 32, 16, sprite=image.Image("sys_updownbuttons_small.png", 32, 16),
                                  on_frame=2, off_frame=3, parent=self)

        self.children.append(up_arrow)
        self.children.append(down_arrow)

        self.on_change = on_change
        # The current index of the cursor and carat:
        self.cursor_i = 0
        self.carat_i = None
        self.top_line = 0
        self.num_lines = (self.h - 12) // self.font.get_linesize()

    def get_line_pos(self, index, lengths):
        # Given the buffer index and lengths of each screen line (in characters), return the screen line
        # and the screen index of this position.
        line = 0
        for l in lengths:
            if index < l:
                break
            else:
                index -= l
                line += 1
        return line, index

    def get_buffer_index(self, dx, dy, lines):
        pass

    def render(self, flash=False):
        mydest = self.get_rect()
        if flash or (self is my_state.active_widget):
            widget_border_on.render(mydest.inflate(-4, -4))
        else:
            widget_border_off.render(mydest.inflate(-4, -4))

        lines, lengths = wrap_with_records(self.text, self.font, self.w - 12)

        cursor_line, cursor_pos = self.get_line_pos(self.cursor_i, lengths)

        if self.top_line + self.num_lines - 1 > len(lines):
            self.top_line = max(0, len(lines) - self.num_lines)

        textdest = mydest.inflate(-12, -12)
        my_state.screen.set_clip(textdest)

        current_line = self.top_line
        # print(cursor_countdown, cursor_line, cursor_pos)
        for l in lines[self.top_line:self.top_line + self.num_lines]:
            img = self.font.render(l, True, self.color)
            my_state.screen.blit(img, textdest)
            if self is my_state.active_widget and current_line == cursor_line:
                cdest = textdest.copy()
                cdest.x += self.font.size(l[:cursor_pos])[0] - 4
                self.cursor_image.render(cdest.topleft, (my_state.anim_phase // 5) % 4)
            current_line += 1
            textdest.y += self.font.get_linesize()

        my_state.screen.set_clip(None)

    def _insert(self, new_text):
        for c in list(new_text):
            self.char_list.insert(self.cursor_i, new_text)
            self.cursor_i += 1

    def _builtin_responder(self, ev):
        if my_state.active_widget is self:
            if ev.type == pygame.KEYDOWN:
                if (ev.key == pygame.K_BACKSPACE) and (len(self.char_list) > 0):
                    if self.cursor_i > 0:
                        del self.char_list[self.cursor_i - 1]
                        self.cursor_i -= 1
                        if self.on_change:
                            self.on_change(self, ev)
                elif (ev.key == pygame.K_DELETE) and (len(self.char_list) > 0):
                    if self.cursor_i < len(self.char_list):
                        del self.char_list[self.cursor_i]
                        if self.on_change:
                            self.on_change(self, ev)
                elif ev.key == pygame.K_LEFT:
                    if self.cursor_i > 0:
                        self.cursor_i -= 1
                elif ev.key == pygame.K_RIGHT:
                    if self.cursor_i < len(self.char_list):
                        self.cursor_i += 1
                elif (ev.unicode in self.ALLOWABLE_CHARACTERS) and (len(ev.unicode) > 0):
                    self._insert(ev.unicode)
                    if self.on_change:
                        self.on_change(self, ev)

    def is_kb_selectable(self):
        return True

    def _get_text(self):
        return "".join(self.char_list)

    def _set_text(self, text):
        self.char_list = list(text)
        if self.on_change:
            self.on_change(self, None)

    text = property(_get_text, _set_text)


# Widgets for columns

class ColTextEntryWidget(RowWidget):
    def __init__(self, width, prompt="Enter Text", text='***', color=None, font=None, justify=-1, on_change=None,
                 **kwargs):
        mylabel = LabelWidget(0, 0, width * 2 // 5 - 10, 0, prompt, font=font)
        super().__init__(0, 0, width, mylabel.h + 8, **kwargs)
        self.add_left(mylabel)
        self.my_text_widget = TextEntryWidget(0, 0, width * 3 // 5, mylabel.h + 8, text, color, font, justify,
                                              on_change=on_change, )
        self.add_right(self.my_text_widget)

    def _get_text(self):
        return self.my_text_widget.text

    def _set_text(self, text):
        self.my_text_widget.text = text

    text = property(_get_text, _set_text)

    def quietly_set_text(self, text):
        # Set the text without calling on_change.
        self.my_text_widget.quietly_set_text(text)


class ColDropdownWidget(RowWidget):
    def __init__(self, width, prompt="Choose Option", font=None, justify=-1, on_select=None, **kwargs):
        mylabel = LabelWidget(0, 0, width * 2 // 5 - 10, 0, prompt, font=font)
        super().__init__(0, 0, width, mylabel.h + 8, **kwargs)
        self.add_left(mylabel)
        self.my_menu_widget = DropdownWidget(0, 0, width * 3 // 5, mylabel.h + 8, font=font, justify=justify,
                                             on_select=on_select)
        self.add_right(self.my_menu_widget)

    def add_item(self, msg, value, desc=None):
        self.my_menu_widget.add_item(msg, value, desc)

    @property
    def value(self):
        return self.my_menu_widget.value



import collections
import math

from .isosm_config import NOT_ALL_FLAGS, FLIPPED_VERTICALLY_FLAG, FLIPPED_HORIZONTALLY_FLAG, SCROLL_STEP
from ... import _hub
from ... import event
from ...compo import vscreen as core

EngineEvTypes = event.EngineEvTypes

# --------------------------------------------
# optimization
_gl_halftile_w = _gl_halftile_h = None
_buffer_relx = dict()
_buffer_rely = dict()


def relative_x(x, y):
    """Return the relative x position of this tile, ignoring offset."""
    try:
        return _buffer_relx[x][y]
    except KeyError:
        if x not in _buffer_relx:
            _buffer_relx[x] = dict()
        _buffer_relx[x][y] = int(x * _gl_halftile_w - y * _gl_halftile_w)
        return _buffer_relx[x][y]


def relative_y(x, y):
    """Return the relative y position of this tile, ignoring offset."""
    try:
        return _buffer_rely[x][y]
    except KeyError:
        if x not in _buffer_rely:
            _buffer_rely[x] = dict()
        _buffer_rely[x][y] = int(y * _gl_halftile_h + x * _gl_halftile_h)
        return _buffer_rely[x][y]


def rel_set_info(halftile_w, halftile_h):
    global _gl_halftile_w, _gl_halftile_h, _buffer_relx, _buffer_rely
    _gl_halftile_w = float(halftile_w)
    _gl_halftile_h = float(halftile_h)
    _buffer_relx.clear()
    _buffer_rely.clear()


# ----------------------------------------------


class IsometricMapViewer(event.EventReceiver):
    MEGAOPTIM = False
    MOUSEMOTION_CONST = None
    SOLID_COLOR = 'navyblue'
    FLOORS = {}
    FLOOR_MAN_OFFSET = {
        0: (1617 + 305, 240 - 128),  # what offset one needs to apply so the floor img is aligned, map 0 (city.png)
        1: (522 + 300, 100 - 92)
    }

    def __init__(self, isometric_map, screen, postfx=None, cursor=None,
                 left_scroll_key=None, right_scroll_key=None, up_scroll_key=None, down_scroll_key=None):
        super().__init__()

        self.MOUSEMOTION_CONST = _hub.pygame.MOUSEMOTION
        self.animated_wallpaper = False
        self.block_wallpaper = False

        self.isometric_map = isometric_map
        self.FLOORS = {
            0: _hub.pygame.image.load('assets/city.png'),
            1: _hub.pygame.image.load('assets/casino.png')
        }
        self.manoffset = self.FLOOR_MAN_OFFSET[0]
        self.scrollable_floor = self.FLOORS[0]

        self.scrollable_floor.set_colorkey((255, 0, 255))

        self.screen = screen
        self.mid = (
            self.screen.get_width() // 2, self.screen.get_height() // 2
        )
        # self.prevx = float('NaN')
        # self.prevy = float('NaN')

        self.floor_rect = _hub.pygame.Rect(0, 0, self.screen.get_width(), self.screen.get_height())

        if isometric_map.wallpaper:
            self.screen_rect = self.screen.get_rect()
            wp_width, wp_height = self.isometric_map.wallpaper.get_size()
            self.wp_temp_rect = _hub.pygame.Rect(0, 0, wp_width, wp_height)
            self.wp_grid_w = self.screen_rect.w // wp_width + 1
            self.wp_grid_h = self.screen_rect.h // wp_height + 1

        # The focus is defined by map coordinates, so a lot of the back and forth between screen and map coords
        # can be cut.
        self._focus_x = 0
        self._focus_y = 0
        self._focused_object_x0 = 0
        self._focused_object_y0 = 0
        self._focused_object = None

        self.phase = 0

        self.tile_width = isometric_map.tile_width
        self.tile_height = isometric_map.tile_height

        self.half_tile_width = isometric_map.tile_width // 2
        self.half_tile_height = isometric_map.tile_height // 2
        rel_set_info(
            self.half_tile_width, self.half_tile_height
        )
        self.visible_area = None
        self._init_visible_area_init(self.screen)

        # _mouse_tile contains the actual tile the mouse is hovering over. However, in most cases what we really want
        # is the location of the mouse cursor. Time to make a property!
        self._mouse_tile = (-1, -1)
        self.postfx = postfx
        self.cursor = cursor
        self.left_scroll_key = left_scroll_key
        self.right_scroll_key = right_scroll_key
        self.up_scroll_key = up_scroll_key
        self.down_scroll_key = down_scroll_key
        self.debug_sprite = None
        self.lastmousepos = None

        # util for the drawing of tiles
        self.line_cache = list()
        self.objgroup_contents = dict()
        self.objgroup_modified_mappos = dict()
        self._camera_updated_this_frame = False

    def _check_mouse_scroll(self, screen_area, mouse_x, mouse_y):
        # Check for map scrolling, depending on mouse position.
        if not self._camera_updated_this_frame:
            if mouse_x < 20:
                dx = -SCROLL_STEP
            elif mouse_x > (screen_area.right - 20):
                dx = +SCROLL_STEP
            else:
                dx = 0

            if mouse_y < 20:
                dy = -SCROLL_STEP
            elif mouse_y > (screen_area.bottom - 20):
                dy = +SCROLL_STEP
            else:
                dy = 0

            if dx or dy:
                self._update_camera(dx, dy)

    def _get_horizontal_line(self, x0, y0, line_number):
        x = x0 + line_number // 2
        y = y0 + (line_number + 1) // 2

        _, py = self.screen_coords(x, y)
        if py > self.visible_area.bottom:
            return None

        mylist = list()
        while self.screen_coords(x, y)[0] < self.visible_area.right:
            if self.isometric_map.on_the_map(x, y):
                mylist.append((x, y))
            x += 1
            y -= 1
        return mylist

    def _init_visible_area_init(self, scr):
        # The visible area describes the region of the map we need to draw
        w, h = scr.get_size()
        # below: x, y do not matter actually
        self.visible_area = _hub.pygame.Rect(0, 0, w + self.half_tile_width, h + 4 * self.tile_height)

    def _model_depth(self, model):
        return relative_y(model.x, model.y)

    def _update_camera(self, dx, dy):
        # If the mouse and the arrow keys conflict, only one of them should win.
        if self._camera_updated_this_frame:
            return
        self._focus_x, self._focus_y = self.isometric_map.clamp_pos([self._focus_x + dx, self._focus_y + dy])

    def _paint_all(self):
        # TODO idea 1 optimize such as the ground layer is blit as one single image, we just move a rect to cut
        # properly from the very large image stored in memory

        #
        # TOM REMARK: but i cannot use this optimization right now, it makes
        # the floor "wiggle" compared to buildings or portals :(

        # TODO maybe?
        #  idea 2 optimize such as if player+cursor are not moving, all graphic elements are saved onto a "texture"
        #  and we just blit this texture instead of computing new stuff

        # Prep the screen.
        wpok = False
        if not self.block_wallpaper:
            if self.isometric_map.wallpaper is not None:
                self.fill_wallpaper()
                wpok = True
        if not wpok:
            self.screen.fill(self.SOLID_COLOR)

        # Check the map scrolling.
        self._camera_updated_this_frame = False
        if self._focused_object:
            if self._focused_object.x != self._focused_object_x0 or self._focused_object.y != self._focused_object_y0:
                self.focus(self._focused_object.x, self._focused_object.y)
                self._focused_object_x0 = self._focused_object.x
                self._focused_object_y0 = self._focused_object.y

        # Disabling mouse scrolling because the right and bottom edges of the screen don't seem to be working.
        #     if self.lastmousepos:
        #         self._check_mouse_scroll(self.visible_area, *self.lastmousepos)

        # ------------------------------------- check this out <<<
        # --- OPTIM blit large image for the ground level (1/3)---
        if self.MEGAOPTIM:
            # a = relative_x(self._focused_object.x, self._focused_object.y)
            # b = relative_y(self._focused_object.x, self._focused_object.y)  # - self.mid[0],  self._focused_object.y-self.mid[1]
            a, b = self.screen_offset()
            self.floor_rect.topleft = self.manoffset[0] - a, self.manoffset[1] - b
        # ------------------------------------- check this out <<<
        # --- OPTIM blit large image for the ground level (2/3)---
            self.screen.blit(self.scrollable_floor, (0, 0), area=self.floor_rect)

        # Record all of the objectgroup contents for display when their tile comes up
        # Also, clamp all object positions. If this is an infinite scrolling map, objects can move off one side to the
        # map to the other. However, in their data, we want the objects to stay within the bounds of the map.
        for k, v in self.isometric_map.objectgroups.items():
            self.objgroup_contents[k] = collections.defaultdict(list)
            for ob in v.contents:
                ob.x, ob.y = self.isometric_map.clamp_pos((ob.x, ob.y))
                sx, sy = self.screen_coords(ob.x, ob.y, v.offsetx, v.offsety)
                mx, my = self.map_x(sx, sy, return_int=False), self.map_y(sx, sy, return_int=False)
                obkey = self.isometric_map.clamp_pos_int((mx, my))
                self.objgroup_contents[k][obkey].append(ob)
                # Also save the mofidied map pos, which will come in handy later.
                self.objgroup_modified_mappos[ob] = (mx, my)

        x, y = self.map_x(0, 0) - 2, self.map_y(0, 0) - 1
        x0, y0 = x, y
        painting_tiles = True
        line_number = 1

        while painting_tiles:
            # In order to allow smooth sub-tile movement of stuff, we have
            # to draw everything in a particular order.
            nuline = self._get_horizontal_line(x0, y0, line_number)
            self.line_cache.append(nuline)
            current_y_offset = self.isometric_map.layers[0].offsety
            current_line = len(self.line_cache) - 1

            for layer_num, layer in enumerate(self.isometric_map.layers):
                # ------------------------------------- check this out <<<
                # --- OPTIM blit large image for the ground level (3/3)---
                if self.MEGAOPTIM:
                    if layer_num == 0:
                        continue

                if current_line >= 0:
                    if current_line > 1 and layer in self.objgroup_contents and self.line_cache[current_line - 1]:
                        # After drawing the terrain last time, draw any objects in the previous cell.
                        for x, y in self.line_cache[current_line - 1]:
                            if self.cursor:
                                if self.cursor.layer_name == layer.name and x == self.cursor.x and y == self.cursor.y:
                                    self.cursor.render(self)

                            ox, oy = x % self.isometric_map.width, y % self.isometric_map.height
                            if (ox, oy) in self.objgroup_contents[layer]:
                                self.objgroup_contents[layer][(ox, oy)].sort(key=self._model_depth)
                                for ob in self.objgroup_contents[layer][(ox, oy)]:
                                    # The following bit of math makes sure that the clamped object positions will
                                    # print at the correct screen positions. Otherwise, if the player/focus is on the
                                    # other side of the map seam, this object will be printed in the wrong position.
                                    mmx, mmy = self.objgroup_modified_mappos[ob]
                                    fx = x + math.modf(mmx)[0]
                                    fy = y + math.modf(mmy)[0]
                                    sx, sy = self.screen_coords(
                                        fx, fy,
                                        self.isometric_map.objectgroups[layer].offsetx,
                                        self.isometric_map.objectgroups[layer].offsety
                                    )
                                    ob(self.screen, sx, sy, self.isometric_map)

                    if self.line_cache[current_line]:
                        for x, y in self.line_cache[current_line]:
                            gid = layer[x, y]
                            tile_id = gid & NOT_ALL_FLAGS
                            if tile_id > 0:
                                my_tile = self.isometric_map.tilesets[tile_id]

                                # Note that x,y refer to IsometricMapObject coordinates, and so 0,0 points at the
                                # top of the "ground" level of a tile. So, we adjust sy before sending the coords
                                # to the printer, so it is pointing at the bottom corner of the tile instead.
                                sx, sy = self.screen_coords(x, y)
                                my_tile.paint_tile(
                                    self.screen, sx, sy + layer.offsety + self.isometric_map.tile_height,
                                                     gid & FLIPPED_HORIZONTALLY_FLAG,
                                                     gid & FLIPPED_VERTICALLY_FLAG)

                    elif self.line_cache[current_line] is None:  # and layer == self.isometric_map.layers[-1]:
                        painting_tiles = False
                else:
                    break
                if layer.offsety < current_y_offset:
                    current_line -= 1
                    current_y_offset = layer.offsety
            line_number += 1

        self.phase = (self.phase + 1) % 640

        del self.line_cache[:]
        self.objgroup_contents.clear()
        self.objgroup_modified_mappos.clear()

    def fill_wallpaper(self):
        for x in range(-1, self.wp_grid_w):
            self.wp_temp_rect.x = self.screen_rect.x + x * self.wp_temp_rect.w
            for y in range(-1, self.wp_grid_h):
                self.wp_temp_rect.y = self.screen_rect.y + y * self.wp_temp_rect.h
                if self.animated_wallpaper:
                    self.wp_temp_rect.y += self.phase

                self.screen.blit(self.isometric_map.wallpaper, self.wp_temp_rect)

    def focus(self, x, y):
        """Move the camera to point at the requested map tile. x,y can be ints or floats."""
        self._focus_x, self._focus_y = self.isometric_map.clamp_pos((x, y))

        # self.floor_rect.x = a
        # self.floor_rect.y = b

    def map_x(self, sx, sy, return_int=True):
        """Return the map x row for the given screen coordinates."""
        x_off, y_off = self.screen_offset()

        # We're going to use the relative coordinates of the tiles instead of the screen coordinates.
        rx = sx - x_off
        ry = sy - y_off

        # Calculate the x position of map_x tile 0 at ry.
        ox = float(-ry * self.half_tile_width) / float(self.half_tile_height)

        # Now that we have that x origin, we can determine this screen position's x coordinate by dividing by the
        # tile width. Fantastic.
        if return_int:
            # Because of the way Python handles division, we need to apply a little nudge right here.
            if rx - ox < 0:
                ox += self.tile_width
            return int(math.floor((rx - ox) / self.tile_width))
        else:
            return (rx - ox) / self.tile_width

    def map_y(self, sx, sy, return_int=True):
        """Return the map y row for the given screen coordinates."""
        x_off, y_off = self.screen_offset()

        # We're going to use the relative coordinates of the tiles instead of the screen coordinates.
        rx = sx - x_off
        ry = sy - y_off

        oy = float(rx * self.half_tile_height) / float(self.half_tile_width)

        # Now that we have that x origin, we can determine this screen position's x coordinate by dividing by the
        # tile width. Fantastic.
        if return_int:
            # Because of the way Python handles division, we need to apply a little nudge right here.
            if ry - oy < 0:
                oy += self.tile_height
            return int(math.floor((ry - oy) / self.tile_height))
        else:
            return (ry - oy) / self.tile_height

    @property
    def mouse_tile(self):
        if self.cursor:
            return self.cursor.x, self.cursor.y
        else:
            return self._mouse_tile

    def proc_event(self, ev, source=None):
        if ev.type == EngineEvTypes.PAINT:
            self._paint_all()

        elif ev.type == self.MOUSEMOTION_CONST:
            mouse_x, mouse_y = core.proj_to_vscreen(ev.pos)
            self.lastmousepos = (mouse_x, mouse_y)
            self._mouse_tile = (self.map_x(mouse_x, mouse_y), self.map_y(mouse_x, mouse_y))
            if self.cursor:
                self.cursor.update(self, ev)

    def screen_coords(self, x, y, extra_x_offset=0, extra_y_offset=0):
        x_off, y_off = self.screen_offset()
        return relative_x(x, y) + x_off + extra_x_offset, relative_y(x, y) + y_off + extra_y_offset

    def screen_offset(self):
        return self.mid[0] - relative_x(self._focus_x, self._focus_y), \
               self.mid[1] - relative_y(self._focus_x, self._focus_y)

    def set_focused_object(self, fo):
        if fo:
            self._focused_object = fo
            self.focus(fo.x, fo.y)
        else:
            self._focused_object = None

    def switch_map(self, isometric_map):
        # ---------- MEGAOPTIM
        if self.MEGAOPTIM:
            if isometric_map.mapname == 'city':
                self.scrollable_floor = self.FLOORS[0]
                self.manoffset = self.FLOOR_MAN_OFFSET[0]
            elif isometric_map.mapname == 'casino':
                self.scrollable_floor = self.FLOORS[1]
                self.manoffset = self.FLOOR_MAN_OFFSET[1]

        self.isometric_map = isometric_map
        self.tile_width = isometric_map.tile_width
        self.tile_height = isometric_map.tile_height

        self.half_tile_width = isometric_map.tile_width // 2
        self.half_tile_height = isometric_map.tile_height // 2
        rel_set_info(
            self.half_tile_width, self.half_tile_height
        )

        if self._focused_object:
            fo = self._focused_object
            self.focus(fo.x, fo.y)
        else:
            self._focus_x = 0
            self._focus_y = 0



"""
Author: wkta-tom
 | thomas@gaudia-tech.com

This (very strange?) file is here only to describe/identify
a subset of pygame, contrary to appearances
creating this file is useful so Kengi becomes "connectable" to other software
components for example to KataSDK, or to KtgVM...

In other words Kengi does not use 100% of pygame capabilities
and it's better this way for many reasons I don't explain here.
"""


class PygameDrawIface:

    @staticmethod
    def circle(surface, color, pos, radius):
        pass

    @staticmethod
    def rect(surface, bidule):
        pass


class PygameMathIface:

    def Vector2(self, **args):
        pass


class PygameIface:

    draw = PygameDrawIface
    math = PygameMathIface

    KEYDOWN = -1
    KEYUP = 2
    K_LEFT = 3

    @staticmethod
    def Color(rgb):
        pass



import collections

from .isosm_config import NOT_ALL_FLAGS, FLIPPED_VERTICALLY_FLAG, FLIPPED_HORIZONTALLY_FLAG, SCROLL_STEP
from ... import _hub
from ... import event
from ...compo import vscreen


pygame = _hub.pygame
EngineEvTypes = event.EngineEvTypes


class IsometricMapViewer0(event.EventReceiver):
    def __init__(self, isometric_map, screen, postfx=None, cursor=None,
                 left_scroll_key=None, right_scroll_key=None, up_scroll_key=None, down_scroll_key=None):
        super().__init__()
        self.is_drawing = True
        self.isometric_map = isometric_map
        self.screen = screen

        # The focus is defined by map coordinates, so a lot of the back and forth between screen and map coords
        # can be cut.
        self._focus_x = 0
        self._focus_y = 0
        self._focused_object_x0 = 0
        self._focused_object_y0 = 0
        self.phase = 0

        self.tile_width = isometric_map.tile_width
        self.tile_height = isometric_map.tile_height
        self.half_tile_width = isometric_map.tile_width // 2
        self.half_tile_height = isometric_map.tile_height // 2

        # _mouse_tile contains the actual tile the mouse is hovering over. However, in most cases what we really want
        # is the location of the mouse cursor. Time to make a property!
        self._mouse_tile = (-1, -1)
        self.postfx = postfx
        self.cursor = cursor
        self.left_scroll_key = left_scroll_key
        self.right_scroll_key = right_scroll_key
        self.up_scroll_key = up_scroll_key
        self.down_scroll_key = down_scroll_key
        self._focused_object = None
        self.debug_sprite = None
        self.lastmousepos = None
        self.visible_area = None
        self.animated_wp = False

        # util for the drawing of tiles
        self.line_cache = list()
        self.objgroup_contents = dict()
        self.objgroup_modified_mappos = dict()

    def _check_mouse_scroll(self, screen_area, mouse_x, mouse_y):
        # Check for map scrolling, depending on mouse position.
        if mouse_x < 20:
            dx = SCROLL_STEP
        elif mouse_x > (screen_area.right - 20):
            dx = -SCROLL_STEP
        else:
            dx = 0

        if mouse_y < 20:
            dy = SCROLL_STEP
        elif mouse_y > (screen_area.bottom - 20):
            dy = -SCROLL_STEP
        else:
            dy = 0

        if dx or dy:
            self._update_camera(dx, dy)

    def _check_origin(self):
        """Make sure the offset point is within map boundaries."""
        mx = self.map_x(self.screen.get_width() // 2, self.screen.get_height() // 2)
        my = self.map_y(self.screen.get_width() // 2, self.screen.get_height() // 2)

        if not self.isometric_map.on_the_map(mx, my):
            if mx < 0:
                mx = 0
            elif mx >= self.isometric_map.width:
                mx = self.isometric_map.width - 1
            if my < 0:
                my = 0
            elif my >= self.isometric_map.height:
                my = self.isometric_map.height - 1
            self.focus(mx, my)

    def _default_offsets_case(self, a, b):
        if a is None:
            a = self.x_off
        if b is None:
            b = self.y_off
        return a, b

    def _get_horizontal_line(self, x0, y0, line_number, visible_area):
        mylist = list()
        x = x0 + line_number // 2
        y = y0 + (line_number + 1) // 2

        if self.relative_y(x, y) + self.y_off > visible_area.bottom:
            return None

        while self.relative_x(x - 1, y - 1) + self.x_off < visible_area.right:
            if self.isometric_map.on_the_map(x, y):
                mylist.append((x, y))
            x += 1
            y -= 1
        return mylist

    def _init_visible_area_init(self, scr):
        # The visible area describes the region of the map we need to draw.
        # It is bigger than the physical screen
        # because we probably have to draw cells that are not fully on the map.
        self.visible_area = scr.get_rect()

        # - temp disabled inflate op. (web ctx issues)
        # visible_area.inflate_ip(self.tile_width, self.isometric_map.tile_height)
        # incremv = self.isometric_map.tile_height + self.half_tile_height - self.isometric_map.layers[-1].offsety
        # visible_area.h += incremv
        # BUT
        # inflate replaced by the followin hack
        self.visible_area.x += 64
        self.visible_area.y -= 64
        self.visible_area.w += 64
        self.visible_area.h += 64 * 2

    def _model_depth(self, model):
        return self.relative_y(model.x, model.y)

    def _update_camera(self, dx, dy):
        # If the mouse and the arrow keys conflict, only one of them should win.
        if self.camera_updated_this_frame:
            return

        nu_x_off = self.x_off + dx
        nu_y_off = self.y_off + dy

        mx = self.map_x(self.screen.get_width() // 2, self.screen.get_height() // 2, nu_x_off, nu_y_off)
        my = self.map_y(self.screen.get_width() // 2, self.screen.get_height() // 2, nu_x_off, nu_y_off)

        if self.isometric_map.on_the_map(mx, my):
            self.x_off = nu_x_off
            self.y_off = nu_y_off
            self.camera_updated_this_frame = True

    def _paint_all(self):
        self.camera_updated_this_frame = False
        if self._focused_object and (self._focused_object_x0 != self._focused_object.x or
                                     self._focused_object_y0 != self._focused_object.y):
            self.focus(self._focused_object.x, self._focused_object.y)
            self._focused_object_x0 = self._focused_object.x
            self._focused_object_y0 = self._focused_object.y
        # hack disable temporarely the scroll via mouse.To enable-> uncomment 2 lines below
        # else:
        #    self._check_mouse_scroll(screen_area, mouse_x, mouse_y)

        x, y = self.map_x(0, 0) - 2, self.map_y(0, 0) - 1
        x0, y0 = x, y
        painting_tiles = True
        line_number = 1

        del self.line_cache[:]
        self.objgroup_contents.clear()

        # Record all of the objectgroup contents for display when their tile comes up
        for k, v in self.isometric_map.objectgroups.items():
            self.objgroup_contents[k] = collections.defaultdict(list)
            for ob in v.contents:
                sx, sy = self.screen_coords(ob.x, ob.y, k.offsetx + v.offsetx, k.offsety + v.offsety)
                obkey = (self.map_x(sx, sy, return_int=True), self.map_y(sx, sy, return_int=True))
                self.objgroup_contents[k][obkey].append(ob)

        while painting_tiles:
            # In order to allow smooth sub-tile movement of stuff, we have
            # to draw everything in a particular order.
            nuline = self._get_horizontal_line(x0, y0, line_number, self.visible_area)
            self.line_cache.append(nuline)
            current_y_offset = self.isometric_map.layers[0].offsety
            current_line = len(self.line_cache) - 1

            for layer_num, layer in enumerate(self.isometric_map.layers):
                if current_line >= 0:
                    if self.line_cache[current_line]:
                        for x, y in self.line_cache[current_line]:
                            gid = layer[x, y]
                            tile_id = gid & NOT_ALL_FLAGS
                            if tile_id > 0:
                                my_tile = self.isometric_map.tilesets[tile_id]

                                sx, sy = self.screen_coords(x, y)
                                my_tile.paint_tile(self.screen, sx, sy + layer.offsety, gid & FLIPPED_HORIZONTALLY_FLAG,
                                        gid & FLIPPED_VERTICALLY_FLAG)

                            if self.cursor:
                                if self.cursor.layer_name == layer.name and x == self.cursor.x and y == self.cursor.y:
                                    self.cursor.render(self)

                    if current_line > 1 and layer in self.objgroup_contents and self.line_cache[current_line - 1]:
                        # After drawing the terrain, draw any objects in the previous cell.
                        for x, y in self.line_cache[current_line - 1]:

                            if (x, y) in self.objgroup_contents[layer]:
                                self.objgroup_contents[layer][(x, y)].sort(key=self._model_depth)
                                for ob in self.objgroup_contents[layer][(x, y)]:
                                    sx, sy = self.screen_coords(
                                        ob.x, ob.y,
                                        layer.offsetx + self.isometric_map.objectgroups[layer].offsetx,
                                        layer.offsety + self.isometric_map.objectgroups[layer].offsety
                                    )
                                    ob(self.screen, sx, sy, self.isometric_map)

                    elif self.line_cache[current_line] is None and layer == self.isometric_map.layers[-1]:
                        painting_tiles = False
                else:
                    break
                if layer.offsety < current_y_offset:
                    current_line -= 1
                    current_y_offset = layer.offsety
            line_number += 1

        self.phase = (self.phase + 1) % 640

    def fill_wallpaper(self):
        screen_rect = self.screen.get_rect()
        wp_width, wp_height = self.isometric_map.wallpaper.get_size()
        grid_w = screen_rect.w // wp_width + 1
        grid_h = screen_rect.h // wp_height + 1
        my_rect = pygame.Rect(0, 0, wp_width, wp_height)

        for x in range(-1, grid_w):
            my_rect.x = screen_rect.x + x * wp_width
            for y in range(-1, grid_h):
                my_rect.y = screen_rect.y + y * wp_height
                if self.animated_wp:
                    my_rect.y += self.phase
                self.screen.blit(self.isometric_map.wallpaper, my_rect)

    def focus(self, x, y):
        """Move the camera to point at the requested map tile. x,y can be ints or floats."""
        if self.isometric_map.on_the_map(int(x+0.99), int(y+0.99)) and not self.camera_updated_this_frame:
            self.x_off = self.screen.get_width() // 2 - self.relative_x(x, y)
            self.y_off = self.screen.get_height() // 2 - self.relative_y(x, y) + self.tile_height
            self.camera_updated_this_frame = True

    def map_x(self, sx, sy, xoffset_override=None, yoffset_override=None, return_int=True):
        """Return the map x row for the given screen coordinates."""
        x_off, y_off = self._default_offsets_case(xoffset_override, yoffset_override)

        # I was having a lot of trouble with this function, I think because GearHead coordinates use the top left
        # of a square 64x64px cell and for this viewer the map coordinates refer to the midbottom of an arbitrarily
        # sized image. So I broke out some paper and rederived the equations from scratch.

        # We're going to use the relative coordinates of the tiles instead of the screen coordinates.
        rx = sx - x_off
        ry = sy - y_off

        return self.static_map_x(rx, ry, self.tile_width, self.tile_height, self.half_tile_width, self.half_tile_height,
                                 return_int=return_int)

    def map_y(self, sx, sy, xoffset_override=None, yoffset_override=None, return_int=True):
        """Return the map y row for the given screen coordinates."""
        x_off, y_off = self._default_offsets_case(xoffset_override, yoffset_override)

        # We're going to use the relative coordinates of the tiles instead of the screen coordinates.
        rx = sx - x_off
        ry = sy - y_off

        return self.static_map_y(rx, ry, self.tile_width, self.tile_height, self.half_tile_width, self.half_tile_height,
                                 return_int=return_int)

    @property
    def mouse_tile(self):
        if self.cursor:
            return self.cursor.x, self.cursor.y
        else:
            return self._mouse_tile

    def pause_draw(self):
        self.is_drawing = False

    def proc_event(self, ev, source=None):
        if not self.is_drawing:
            return
        if ev.type == EngineEvTypes.PAINT:
            if self.visible_area is None:
                self._init_visible_area_init(ev.screen)
            ev.screen.fill('black')
            self._paint_all()
            return
        if ev.type == pygame.MOUSEMOTION:
            mouse_x, mouse_y = vscreen.proj_to_vscreen(ev.pos)
            self.lastmousepos = (mouse_x, mouse_y)
            self._mouse_tile = (self.map_x(mouse_x, mouse_y), self.map_y(mouse_x, mouse_y))
            if self.cursor:
                self.cursor.update(self, ev)

    def relative_x(self, x, y):
        """Return the relative x position of this tile, ignoring offset."""
        return (x * self.half_tile_width) - (y * self.half_tile_width)

    def relative_y(self, x, y):
        """Return the relative y position of this tile, ignoring offset."""
        return (y * self.half_tile_height) + (x * self.half_tile_height)

    def resume_draw(self):
        self.is_drawing = True

    def screen_coords(self, x, y, extra_x_offset=0, extra_y_offset=0):
        return (self.relative_x(x - 1, y - 1) + self.x_off + extra_x_offset,
                self.relative_y(x - 1, y - 1) + self.y_off + extra_y_offset)

    def scroll_to(self, directioncode):
        """
        :param directioncode: 0 - 3 such as 0 north, 1 east, 2 south, 3 west
        """
        dx = dy = 0
        if 0 == directioncode:
            dy = SCROLL_STEP
        elif 2 == directioncode:
            dy = -SCROLL_STEP
        elif 3 == directioncode:
            dx = SCROLL_STEP
        elif 1 == directioncode:
            dx = -SCROLL_STEP
        else:
            raise ValueError('invalid directioncode: ', directioncode)
        self._update_camera(dx, dy)

    def set_focused_object(self, fo):
        if fo:
            self._focused_object = fo
            self._focused_object_x0 = fo.x
            self._focused_object_y0 = fo.y
        else:
            self._focused_object = None

    @staticmethod
    def static_map_x(rx, ry, tile_width, tile_height, half_tile_width, half_tile_height, return_int=True):
        # Return the map coordinates for the relative_x, relative_y coordinates. All x,y offsets- including both
        # the view offset and the layer offset- should already have been applied. This method is needed for calculating
        # the layer coords of objects imported from Tiled, which have pixel coords.
        #
        # Calculate the x position of map_x tile -1 at ry. There is no tile -1, but this is the origin from which we
        # measure everything.
        ox = float(-ry * half_tile_width) / half_tile_height - tile_width

        # Now that we have that x origin, we can determine this screen position's x coordinate by dividing by the
        # tile width. Fantastic.
        if return_int:
            # Because of the way Python handles division, we need to apply a little nudge right here.
            if rx - ox < 0:
                ox += tile_width
            return int((rx - ox) / tile_width) + 1
        else:
            return (rx - ox) / tile_width + 1

    @staticmethod
    def static_map_y(rx, ry, tile_width, tile_height, half_tile_width, half_tile_height, return_int=True):
        # Return the map coordinates for the relative_x, relative_y coordinates. All x,y offsets- including both
        # the view offset and the layer offset- should already have been applied. This method is needed for calculating
        # the layer coords of objects imported from Tiled, which have pixel coords.
        #
        # Calculate the x position of map_x tile -1 at ry. There is no tile -1, but this is the origin from which we
        # measure everything.
        oy = float(rx * half_tile_height) / half_tile_width - tile_height

        # Now that we have that x origin, we can determine this screen position's x coordinate by dividing by the
        # tile width. Fantastic.
        if return_int:
            # Because of the way Python handles division, we need to apply a little nudge right here.
            if ry - oy < 0:
                oy += tile_height
            return int((ry - oy) / tile_height) + 1
        else:
            return (ry - oy) / tile_height + 1

    def switch_map(self, isometric_map):
        self.isometric_map = isometric_map
        self.tile_width = isometric_map.tile_width
        self.tile_height = isometric_map.tile_height
        self.half_tile_width = isometric_map.tile_width // 2
        self.half_tile_height = isometric_map.tile_height // 2
        self._check_origin()



import random
import weakref

from . import frects
from . import image
from .image import TEXT_COLOR, truncline, render_text
from ... import _hub
from ...compo import vscreen


pygame = _hub.pygame


class KeyObject(object):
    """A catcher for multiple inheritence. Subclass this instead of object if
       you're going to use multiple inheritence, so that erroneous keywords
       will get caught and identified."""

    def __init__(self, **keywords):
        for k, i in keywords.items():
            print("WARNING: KeyObject got parameters {}={}".format(k, i))


class SingletonMeta(type):
    def __str__(cls):
        return cls.name


class Singleton(object, metaclass=SingletonMeta):
    """For rules constants that don't need to be instanced."""
    name = "Singleton"
    def __init__(self):
        raise NotImplementedError("Singleton can't be instantiated.")


class Border(object):
    def __init__(self, border_width=16, tex_width=32, border_name="", tex_name="", padding=16, tl=0, tr=0, bl=0, br=0,
                 t=1, b=1, l=2, r=2, transparent=True):
        # tl,tr,bl,br are the top left, top right, bottom left, and bottom right frames
        # Bug: The border must be exactly half as wide as the texture.
        self.border_width = border_width
        self.tex_width = tex_width
        self.border_name = border_name
        self.tex_name = tex_name
        self.border = None
        self.tex = None
        self.padding = padding
        self.tl = tl
        self.tr = tr
        self.bl = bl
        self.br = br
        self.t = t
        self.b = b
        self.l = l
        self.r = r
        self.transparent = transparent

    def render(self, dest, scr=None):
        """Draw this decorative border at dest on screen."""
        if scr is None:
            scr = core.get_screen()

        # We're gonna draw a decorative border to surround the provided area.
        if self.border == None:
            self.border = image.Image(self.border_name, self.border_width, self.border_width)
        if self.tex_name and not self.tex:
            self.tex = image.Image(self.tex_name, self.tex_width, self.tex_width)
            if self.transparent:
                self.tex.bitmap.set_alpha(224)

        # Draw the backdrop.
        if self.tex:
            self.tex.tile(dest.inflate(self.padding, self.padding))

        # Expand the dimensions to their complete size.
        # The method inflate_ip doesn't seem to be working... :(
        fdest = dest.inflate(self.padding, self.padding)

        self.border.render((fdest.x - self.border_width // 2, fdest.y - self.border_width // 2), self.tl)
        self.border.render((fdest.x - self.border_width // 2, fdest.y + fdest.height - self.border_width // 2), self.bl)
        self.border.render((fdest.x + fdest.width - self.border_width // 2, fdest.y - self.border_width // 2), self.tr)
        self.border.render(
            (fdest.x + fdest.width - self.border_width // 2, fdest.y + fdest.height - self.border_width // 2), self.br)

        fdest = dest.inflate(self.padding - self.border_width, self.padding + self.border_width)
        scr.set_clip(fdest)
        for x in range(0, fdest.w // self.border_width + 2):
            self.border.render((fdest.x + x * self.border_width, fdest.y), self.t)
            self.border.render((fdest.x + x * self.border_width, fdest.y + fdest.height - self.border_width), self.b)

        fdest = dest.inflate(self.padding + self.border_width, self.padding - self.border_width)
        scr.set_clip(fdest)
        for y in range(0, fdest.h // self.border_width + 2):
            self.border.render((fdest.x, fdest.y + y * self.border_width), self.l)
            self.border.render((fdest.x + fdest.width - self.border_width, fdest.y + y * self.border_width), self.r)
        scr.set_clip(None)


# Monkey Type these definitions to fit your game/assets.
default_border = Border(border_width=8, tex_width=16, border_name="assets/sys_defborder.png", tex_name="assets/sys_defbackground.png",
                        tl=0, tr=3, bl=4, br=5, t=1, b=1, l=2, r=2)
notex_border = Border(border_width=8, border_name="assets/sys_defborder.png", padding=4, tl=0, tr=3, bl=4, br=5, t=1, b=1, l=2,
                      r=2)
# map_border = Border( border_name="sys_mapborder.png", tex_name="sys_maptexture.png", tl=0, tr=1, bl=2, br=3, t=4, b=6, l=7, r=5 )
# gold_border = Border( border_width=8, tex_width=16, border_name="sys_rixsborder.png", tex_name="sys_rixstexture.png", tl=0, tr=3, bl=4, br=5, t=1, b=1, l=2, r=2 )

WHITE = (255, 255, 255)
GREY = (160, 160, 160)

INFO_GREEN = (50, 200, 0)
INFO_HILIGHT = (100, 250, 0)
ENEMY_RED = (250, 50, 0)


class GameState(object):
    def __init__(self, screen=None):
        self.screen = screen
        self.physical_screen = None
        self.view = None
        self.got_quit = False
        self.widgets = list()
        self.widgets_active = True
        self.active_widget_hilight = False
        self._active_widget = None
        self.widget_clicked = False
        self.audio_enabled = True
        self.music = None
        self.music_name = None
        self.music_library = dict()
        self.anim_phase = 0
        self.standing_by = False
        self.notifications = list()

        self.mouse_pos = (0, 0)

    def render_widgets(self):
        if self.widgets:
            for w in self.widgets:
                w.super_render()
        elif self.active_widget_hilight:
            self.active_widget_hilight = False

    def render_notifications(self):
        for n in list(self.notifications):
            n.render()
            if n.is_done():
                self.notifications.remove(n)

    def do_flip(self, show_widgets=True, reset_standing_by=True):
        self.widget_tooltip = None
        if show_widgets:
            self.render_widgets()
        if self.notifications:
            self.render_notifications()
        if self.widget_tooltip:
            x, y = self.mouse_pos
            x += 16
            y += 16
            if x + 200 > self.screen.get_width():
                x -= 200
            myimage = render_text(self.small_font, self.widget_tooltip, 200)
            myrect = myimage.get_rect(topleft=(x, y))
            default_border.render(myrect)
            self.screen.blit(myimage, myrect)
        self.anim_phase = (self.anim_phase + 1) % 6000
        if reset_standing_by:
            self.standing_by = False
        pygame.display.flip()


    def _set_active_widget(self, widj):
        if widj:
            self._active_widget = weakref.ref(widj)
        else:
            self._active_widget = None

    def _get_active_widget(self):
        if self._active_widget:
            return self._active_widget()

    def _del_active_widget(self):
        self._active_widget = None

    active_widget = property(_get_active_widget, _set_active_widget, _del_active_widget)

    def _get_all_kb_selectable_widgets(self, wlist):
        mylist = list()
        for w in wlist:
            if w.active:
                if w.is_kb_selectable():
                    mylist.append(w)
                if w.children:
                    mylist += self._get_all_kb_selectable_widgets(w.children)
        return mylist

    def activate_next_widget(self, backwards=False):
        wlist = self._get_all_kb_selectable_widgets(self.widgets)
        awid = self.active_widget
        if awid and awid in wlist:
            if backwards:
                n = wlist.index(awid) - 1
            else:
                n = wlist.index(awid) + 1
                if n >= len(wlist):
                    n = 0
            self.active_widget = wlist[n]
        elif wlist:
            self.active_widget = wlist[0]

    def resize(self):
        w, h = self.physical_screen.get_size()
        self.screen = pygame.Surface((max(800, 600 * w // h), 600))


INPUT_CURSOR = None
SMALLFONT = None
TINYFONT = None
ITALICFONT = None
BIGFONT = None
ANIMFONT = None
MEDIUMFONT = None
ALTTEXTFONT = None  # Use this instead of MEDIUMFONT when you want to shake things up a bit.
POSTERS = list()

my_state = GameState()
# fix: add ref to screen
my_state.screen = vscreen.get_screen()

# The FPS the rules runs at.
FPS = 30

# Use a timer to control FPS.
TIMEREVENT = pygame.USEREVENT

# Remember whether or not this unit has been initialized, since we don't need
# to initialize it more than once.
INIT_DONE = False


def wrap_with_records(fulltext, font, maxwidth):
    # Do a word wrap, but also return the length of each line including whitespace and newlines.
    done = 0
    wrapped = list()
    line_lengths = list()

    for text in fulltext.splitlines(True):
        done = 0
        while not done:
            nl, done, stext = truncline(text, font, maxwidth)
            wrapped.append(stext.lstrip())
            # wrapped.append(stext)
            line_lengths.append(nl + 1)
            text = text[nl:]
    return wrapped, line_lengths


def wait_event():
    # Wait for input, then return it when it comes.
    ev = pygame.event.wait()

    # Record if a quit event took place
    if ev.type == pygame.QUIT:
        my_state.got_quit = True
    elif ev.type == TIMEREVENT:
        pygame.event.clear(TIMEREVENT)

    # Inform any interested widgets of the event.
    my_state.widget_clicked = False
    if my_state.widgets_active:
        for w in my_state.widgets:
            w.respond_event(ev)

    # If the view has a check_event method, call that.
    if my_state.view and hasattr(my_state.view, "check_event"):
        my_state.view.check_event(ev)

    return ev




def alert(text, font=None):
    if not font:
        font = my_state.medium_font
    # mydest = pygame.Rect( my_state.screen.get_width() // 2 - 200, my_state.screen.get_height()//2 - 100, 400, 200 )
    mytext = render_text(font, text, 400)
    mydest = mytext.get_rect(center=(my_state.screen.get_width() // 2, my_state.screen.get_height() // 2))

    pygame.event.clear([TIMEREVENT, pygame.KEYDOWN])
    while True:
        ev = wait_event()
        if (ev.type == pygame.MOUSEBUTTONUP) or (ev.type == pygame.QUIT) or (ev.type == pygame.KEYDOWN):
            return ev
        elif ev.type == TIMEREVENT:
            if my_state.view:
                my_state.view()
            default_border.render(mydest)
            my_state.screen.blit(mytext, mydest)
            my_state.do_flip()


def alert_display(display_fun):
    pygame.event.clear([TIMEREVENT, pygame.KEYDOWN])
    while True:
        ev = wait_event()
        if (ev.type == pygame.MOUSEBUTTONUP) or (ev.type == pygame.QUIT) or (ev.type == pygame.KEYDOWN):
            break
        elif ev.type == TIMEREVENT:
            if my_state.view:
                my_state.view()
            display_fun()
            my_state.do_flip()


ALLOWABLE_CHARACTERS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ 1234567890()-=_+,.?"'


def input_string(font=None, redrawer=None, prompt="Enter text below", prompt_color=(255, 255, 255),
                 input_color=TEXT_COLOR, border=default_border):
    # Input a string from the user.
    it = []
    keep_going = True

    if not font:
        font = BIGFONT

    myrect = pygame.Rect(my_state.screen.get_width() / 2 - 200, my_state.screen.get_height() / 2 - 32, 400, 64)
    prompt_image = font.render(prompt, True, prompt_color)

    while keep_going:
        ev = wait_event()

        if ev.type == TIMEREVENT:
            if redrawer != None:
                redrawer()
            border.render(myrect)
            mystring = "".join(it)
            myimage = font.render(mystring, True, input_color)
            my_state.screen.blit(prompt_image, (my_state.screen.get_width() / 2 - prompt_image.get_width() / 2,
                                                my_state.screen.get_height() / 2 - prompt_image.get_height() - 2))
            my_state.screen.set_clip(myrect)
            my_state.screen.blit(myimage, (
            my_state.screen.get_width() / 2 - myimage.get_width() / 2, my_state.screen.get_height() / 2))
            INPUT_CURSOR.render(
                (my_state.screen.get_width() / 2 + myimage.get_width() / 2 + 2, my_state.screen.get_height() / 2),
                ( my_state.anim_phase // 3 ) % 4)
            my_state.screen.set_clip(None)
            my_state.do_flip(False)


        elif ev.type == pygame.KEYDOWN:
            if (ev.key == pygame.K_BACKSPACE) and (len(it) > 0):
                del it[-1]
            elif (ev.key == pygame.K_RETURN) or (ev.key == pygame.K_ESCAPE):
                keep_going = False
            elif (ev.unicode in ALLOWABLE_CHARACTERS) and (len(ev.unicode) > 0):
                it.append(ev.unicode)
        elif ev.type == pygame.QUIT:
            keep_going = False
    return "".join(it)


def please_stand_by(caption=None):
    if not my_state.standing_by:
        img = pygame.image.load(random.choice(POSTERS)).convert()
        dest = img.get_rect(center=(my_state.screen.get_width() // 2, my_state.screen.get_height() // 2))
        my_state.screen.fill((0, 0, 0))
        my_state.screen.blit(img, dest)
        if caption:
            mytext = BIGFONT.render(caption, True, TEXT_COLOR)
            dest2 = mytext.get_rect(topleft=(dest.x + 32, dest.y + 32))
            default_border.render(my_state.screen, dest2)
            my_state.screen.blit(mytext, dest2)
        my_state.standing_by = True
        my_state.do_flip(False, reset_standing_by=False)


class BasicNotification(frects.Frect):
    IP_INFLATE = 0
    IP_DISPLAY = 1
    IP_DEFLATE = 2
    IP_DONE = 3

    def __init__(self, text, font=None, dx=16, dy=16, w=256, h=10, anchor=frects.ANCHOR_UPPERLEFT,
                 border=default_border, count=60, **kwargs):
        font = font or my_state.big_font
        w = min(w, font.size(text)[0])
        self.text_bitmap = render_text(font, text, w)
        h = max(h, self.text_bitmap.get_height())
        super().__init__(dx, dy, w, h, anchor, **kwargs)
        self.border = border
        self.count = count
        self._inflation_phase = self.IP_INFLATE
        self._inflation_count = 0
        my_state.notifications.append(self)

    def render(self):
        if self._inflation_phase == self.IP_INFLATE:
            # Inflating
            mydest = self.get_rect()
            mydest.inflate_ip(-(self.w * (5 - self._inflation_count)) // 6,
                              -(self.h * (5 - self._inflation_count)) // 6)
            self.border.render(mydest)
            self._inflation_count += 1
            if self._inflation_count >= 5:
                self._inflation_phase = self.IP_DISPLAY
        elif self._inflation_phase == self.IP_DISPLAY and self.count > 0:
            self.border.render(self.get_rect())
            my_state.screen.blit(self.text_bitmap, self.get_rect())
            self.count -= 1
        else:
            mydest = self.get_rect()
            mydest.inflate_ip(-(self.w * (5 - self._inflation_count)) // 6,
                              -(self.h * (5 - self._inflation_count)) // 6)
            self.border.render(mydest)
            self._inflation_count -= 1
            if self._inflation_count <= 0:
                self._inflation_phase = self.IP_DONE

    def is_done(self):
        return self._inflation_phase == self.IP_DONE


# PG2 Change
# FULLSCREEN_FLAGS = pygame.FULLSCREEN | pygame.SCALED
# WINDOWED_FLAGS = pygame.RESIZABLE | pygame.SCALED
# FULLSCREEN_RES = (800,600)

# -
# FULLSCREEN_FLAGS = pygame.FULLSCREEN
# WINDOWED_FLAGS = pygame.RESIZABLE
# if sys.platform.startswith("linux"):
#     FULLSCREEN_RES = (800, 600)
# else:
#     FULLSCREEN_RES = (0, 0)



import re
from . import _hub


FONT = 'consolas'


def _font(size):
    return _hub.kengi_inj['pygame'].font.SysFont(FONT, size)


def camel_case_format(string_ac_underscores):
    words = [word.capitalize() for word in string_ac_underscores.split('_')]
    return "".join(words)


def drawtext(msg, size=50, color=(255, 255, 255), aliased=False):
    return _font(size).render(msg, aliased, color)


def underscore_format(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



from .Trigger import Trigger

from ... import _hub as inj


pygame = inj.pygame


class SpriteBo:
    """
    on définit un sprite comme un élément graphique connu,
    et déjà présent en mémoire,
    dont on place une copie qq part sur lécran...
    Chaque sprite a sa position propre !
    """

    current_z_index = 0

    loaded_surfaces = {}

    @classmethod
    def link_resource(cls, gfxelt_id, filename):
        # this method is new (2022), it has been introduced so we can remove the use to GfxManager.
        cls.loaded_surfaces[gfxelt_id] = pygame.image.load(filename)

    def __init__(self, gfxelt_id, position2):
        self.pos = tuple(position2)
        self.ref_gfxelt_id = gfxelt_id
        self.z_index = self.__class__.current_z_index
        self.__class__.current_z_index += 1

    def getSurface(self):
        return self.__class__.loaded_surfaces[self.ref_gfxelt_id]  # assuming the link has been done

    def draw(self, surf):
        # self.gfx_manager.pasteGfxTo(self.ref_gfxelt_id, self.pos, surf)
        s = self.getSurface()
        surf.blit(s, self.pos)

    def getSize(self):
        # return self.gfx_manager.getSize(self.ref_gfxelt_id)
        return self.getSurface().get_size()

    def getOriginPt(self):
        # return self.gfx_manager.getOrigin(self.ref_gfxelt_id)
        return 0, 0

    def getPos(self):
        return self.pos

    def setPos(self, nouv_position2):
        self.pos = nouv_position2

    def setZIndex(self, new_val):
        self.z_index = new_val

    def getZIndex(self):
        return self.z_index

    def getGfx(self):
        return self.ref_gfxelt_id

    def setGfx(self, nouv_gfxelt_id):
        self.ref_gfxelt_id = nouv_gfxelt_id


class WidgetBo(SpriteBo):
    # classe qui utilise indirectement (via Sprite)
    # les méthodes statiques de atoms.GfxElementManage
    # pour son affichage

    def __init__(self, gfxelt_id, position2):
        super().__init__(gfxelt_id, position2)
        self.__trigger_behavior = Trigger((0, 0), (32, 32))  # position, taille
        self.hidden = False
        self.debugOff()
        self.autoResizeCollisionBox()
        self.__auto_translate_cbox()

    def setGfx(self, nouvel_id):
        super().setGfx(nouvel_id)

    def setPos(self, position2):
        super().setPos(position2)  # position sprite

        self.__auto_translate_cbox()

    def draw(self, screen):
        if not self.hidden:
            super().draw(screen)
            self.__trigger_behavior.draw(screen)

    # --- méthodes agissant sur le trigger interne
    # ----------------------------------------------
    def autoResizeCollisionBox(self):
        current_img_size = self.getSize()
        self.__trigger_behavior.setSize(current_img_size)

    def contains(self, point2):
        return self.__trigger_behavior.contains(point2)

    def __auto_translate_cbox(self):
        tmp = [0, 0]

        obj_pos = self.getPos()
        tmp[0] += obj_pos[0]
        tmp[1] += obj_pos[1]

        org = self.getOriginPt()
        tmp[0] -= org[0]
        tmp[1] -= org[1]

        self.__trigger_behavior.setPos(tmp)

    def set_active(self, val):
        self.__trigger_behavior.active = val
        if val:
            self.hidden = False
        else:
            self.hidden = True

    def is_active(self):
        return self.__trigger_behavior.active

    def debugOn(self):
        self.__trigger_behavior.set_visible(True)

    def debugOff(self):
        self.__trigger_behavior.set_visible(False)

    def getLabel(self):
        return self.__trigger_behavior.getLabel()

    def setLabel(self, label=None):
        self.__trigger_behavior.setLabel(label)

    def getCollisionBox(self):
        return self.__trigger_behavior.getCollisionBox()

    def setCollisionBox(self, li4_coord):
        self.__trigger_behavior.setCollisionBox(li4_coord)



import sys
import traceback
from katagames_sdk.capsule.engine_ground.BaseGameState import BaseGameState
from katagames_sdk.capsule.bioslike.MicroView import MicroView
from katagames_sdk.capsule.engine_ground.util import underscore_format
from katagames_sdk.capsule.struct.Singleton import Singleton

# import glvars
# import katagames_sdk.engine as kataen
# from app.title_screen.MenuCtrl import MenuCtrl
# from app.title_screen.TitleModel import TitleModel
# from app.title_screen.TitleView import TitleView
# pygame = kataen.import_pygame()

# import glvars
# import katagames_sdk.api as katapi
# import katagames_sdk.engine as kataen
# from app.login.LoginCtrl import LoginCtrl
# from app.login.LoginMod import LoginMod
# from app.login.LoginView import LoginView
# from ev_types import MyEvTypes
# from katagames_sdk.engine import EventReceiver, EngineEvTypes
# pygame = kataen.import_pygame()
from katagames_sdk.capsule.cli_side_api import has_pre_auth, get_credentials


class KataFrameState(BaseGameState):

    def __init__(self, gs_ident, name):
        super().__init__(gs_ident, name)
        self.m = self.v = self.c = None
        self.v2 = None
        self.glvars_module = None

    def enter(self):
        print('BIOS in')
        from katagames_sdk.capsule.bioslike.KataFrameM import KataFrameM
        from katagames_sdk.capsule.bioslike.KataFrameV import KataFrameV
        from katagames_sdk.capsule.bioslike.KataFrameC import KataFrameC

        self.m = KataFrameM()

        # the state can change right away, if login info is found...
        if not has_pre_auth():
            self.v = KataFrameV(self.m)
        else:
            username, player_id = get_credentials()
            print('fetched from session storage: {}, {}'.format(username, player_id))
            self.glvars_module.username = username
            self.glvars_module.acc_id = player_id
            self.v = MicroView()

        self.c = KataFrameC(self.m, self.v)
        self.c.glvars_module = self.glvars_module

        self.resume()

    def release(self):
        print('BIOS out')
        self.pause()
        self.m = self.v = self.c = None

    def pause(self):
        if self.v:
            self.v.turn_off()
        else:
            self.v2.turn_off()
        self.c.turn_off()

    def resume(self):
        self.v.turn_on()
        self.c.turn_on()


@Singleton
class StContainer:
    """
    contient toutes les instances de classes qui dérivent de BaseGameState
    """
    def __init__(self):
        self.__setup_done = False
        self.assoc_id_state_obj = dict()

    def hack_bios_state(self, gs_obj):
        self.assoc_id_state_obj[-1] = gs_obj

    def setup(self, enum_game_states, pymodule, stmapping):
        self.__setup_done = True

        gs_obj = KataFrameState(-1, 'KataFrame')
        gs_obj.glvars_module = pymodule
        self.hack_bios_state(gs_obj)

        # -- initialisation d'après ce qu'on recoit comme enumeration...
        for id_choisi, nom_etat in enum_game_states.inv_map.items():
            nom_cl = nom_etat + 'State'

            if stmapping:  # but = charger en mémoire la classe
                adhoc_cls = stmapping[nom_cl]  # class has been provided

                obj = adhoc_cls(id_choisi, nom_etat)
                self.assoc_id_state_obj[id_choisi] = obj
            else:
                pymodule_name = underscore_format(nom_etat)
                pythonpath = 'app.{}.state'.format(pymodule_name)
                print('StContainer is loading a new game state...')
                try:
                    pymodule = __import__(pythonpath, fromlist=[nom_cl])
                    adhoc_cls = getattr(pymodule, nom_cl)  # class has been retrieved -> ok

                    obj = adhoc_cls(id_choisi, nom_etat)
                    self.assoc_id_state_obj[id_choisi] = obj

                except ImportError as exc:
                    print('adhoc module name(conv. to underscore format)={}'.format(pymodule_name))
                    print('full path for finding class={}'.format(pythonpath))
                    print('target class={}'.format(nom_cl))
                    print('WEB CONTEXT WARNING: make sure you dont have a file named app.py in your project!')
                    sys.stderr.write("Error: failed to import class {} (info= {})\n".format(nom_cl, exc))
                    traceback.print_last()

    def retrieve(self, identifiant):
        """
        :param identifiant: peut-être aussi bien le code (int) que le nom de classe dédiée (e.g. PlayState)
        :return: instance de BaseGameState
        """

        # construction par nom ou identifiant entier
        # TODO rétablir recherche par nom et non par code...

        # if isinstance(identifiant, str):
        #     gamestate_id = None
        #     for num_id, nom in state_listing.items():
        #         if nom == identifiant:
        #             gamestate_id = num_id
        #             break
        #     if gamestate_id is None:
        #         assert 0, "state name not found: " + identifiant
        # else:
        #     gamestate_id = identifiant
        gamestate_id = identifiant
        return self.assoc_id_state_obj[gamestate_id]



from abc import ABCMeta, abstractmethod


class BaseGameState(metaclass=ABCMeta):

    def __init__(self, state_ident):
        self.__state_ident = state_ident
        self.__state_name = self.__class__.__name__

    @abstractmethod
    def enter(self):
        pass

    def get_id(self):
        return self.__state_ident

    def get_name(self):
        return self.__state_name

    def pause(self):
        print(self.__class__.__name__)
        raise AssertionError('not meant to be paused')

    @abstractmethod
    def release(self):
        pass

    def resume(self):
        print(self.__class__.__name__)
        raise AssertionError('not meant to be resumed')



import collections
import glob
import random

from ... import _hub
from ... import event
from ...compo import vscreen as core


pygame = _hub.pygame  # alias to keep on using pygame, easily
Frect = _hub.polarbear.frects.Frect
ANCHOR_CENTER = _hub.polarbear.frects.ANCHOR_CENTER
ANCHOR_UPPERLEFT = _hub.polarbear.frects.ANCHOR_UPPERLEFT
default_border = _hub.polarbear.default_border
render_text = _hub.polarbear.render_text
draw_text = _hub.polarbear.draw_text
wait_event = _hub.polarbear.wait_event
TIMEREVENT = _hub.polarbear.TIMEREVENT
INFO_GREEN = _hub.polarbear.INFO_GREEN
DEFAULT_FONT_SIZE = 11
MENU_ITEM_COLOR = pygame.Color(150, 145, 130)
MENU_SELECT_COLOR = pygame.Color(128, 250, 230)
ReceiverObj = event.EventReceiver
EngineEvTypes = event.EngineEvTypes


class MenuItem(object):
    def __init__(self, msg, value, desc, menu):
        self.value = value
        self.desc = desc
        self.font = menu.font
        self.width = menu.w
        self.justify = -1
        self.menuitem = menu.menuitem
        self.menuselect = menu.menuselect
        self.item_image = None
        self.select_image = None
        self.height = 0
        self.msg = msg

    def _get_msg(self):
        return self._msg

    def _set_msg(self, msg):
        self._msg = msg
        self.item_image = render_text(self.font, self._msg, self.width, justify=self.justify, color=self.menuitem)
        self.select_image = render_text(self.font, self._msg, self.width, justify=self.justify, color=self.menuselect)
        self.height = self.select_image.get_height()

    msg = property(_get_msg, _set_msg)

    SORT_LAYER = 0

    def sort_order(self):
        return self.SORT_LAYER, self._msg

    def __lt__(self, other):
        """ Comparison of menu items done by sort order, as defined above """
        return self.sort_order() < other.sort_order()

    def __str__(self):
        return self._msg

    def render(self, screen, dest, selected=False):
        if selected:
            screen.blit(self.select_image, dest)
        else:
            screen.blit(self.item_image, dest)


# The DescBox is the default MenuDesc. It takes a string stored in the menu
# item and displays it. However, it is not the only menu description possible!
# Any object with a render_desc(menu_item) method will work.
# Also note that the desc associated with each menu item doesn't need to be
# a string- it all depends on the needs of the descobj you're using.

class DescBox(Frect):
    # The DescBox inherits from Frect, since that's basically what it is.
    def __init__(self, menu, dx, dy, w=300, h=100, anchor=ANCHOR_CENTER, border=default_border, justify=-1, font=None,
                 color=None, **kwargs):
        self.menu = menu
        self.border = border
        self.justify = justify
        if not anchor:
            anchor = menu.anchor
        self.font = font or pygame.font.Font("assets/DejaVuSansCondensed-Bold.ttf", DEFAULT_FONT_SIZE)
        self.color = color or INFO_GREEN
        super(DescBox, self).__init__(dx, dy, w, h, anchor, **kwargs)

    def __call__(self, menu_item):
        mydest = self.get_rect()
        if self.border:
            self.border.render(mydest)
        if menu_item and menu_item.desc:
            img = render_text(self.font, menu_item.desc, self.w, justify=self.justify, color=self.color)
            screen.blit(img, mydest)


class Menu(event.CogObj, Frect):  # N.B (tom) it would be better to inherit from CogObj +have a Frect attribute

    def __init__(self, dx, dy, w=300, h=100, anchor=ANCHOR_CENTER, menuitem=MENU_ITEM_COLOR,
                 menuselect=MENU_SELECT_COLOR, border=default_border, predraw=None, font=None, padding=0,
                 item_class=MenuItem):
        # usin multiple inheritance
        event.CogObj.__init__(self)
        Frect.__init__(self, dx, dy, w, h, anchor)

        self.menuitem = menuitem
        self.menuselect = menuselect
        self.border = border
        self.font = font or pygame.font.Font("assets/DejaVuSansCondensed-Bold.ttf", DEFAULT_FONT_SIZE)
        self.more_image = self.font.render("+", True, menuselect)
        self.padding = padding
        self.item_class = item_class

        self.items = []
        self.top_item = 0
        self.selected_item = 0
        self.can_cancel = True
        self.descobj = None
        self.quick_keys = {}

        self._item_rects = collections.OrderedDict()
        self._the_highest_top = 0

        # predraw is a function that
        # redraws/clears the screen before the menu is rendered.
        self.predraw = predraw

        # -> to the model
        self.no_choice_made = True
        self.choice = False

    def add_item(self, msg, value, desc=None):
        item = self.item_class(msg, value, desc, self)
        self.items.append(item)

    def add_descbox(self, x, y, w=30, h=10, justify=-1, font=None, **kwargs):
        self.descobj = DescBox(self, x, y, w, h, border=self.border, justify=justify, font=font or self.font, **kwargs)

    def arrange(self):
        # Set the position of all on-screen menu items.
        mydest = self.get_rect()
        item_num = self.top_item
        self._item_rects.clear()
        y = mydest.top
        while y < mydest.bottom:
            if item_num < len(self.items):
                itemdest = pygame.Rect(mydest.x, y, self.w, self.items[item_num].height)
                # Only add this item to the menu if it fits inside the menu or is the first menu item.
                if itemdest.bottom <= mydest.bottom or not self._item_rects:
                    self._item_rects[item_num] = itemdest
                y += self.items[item_num].height + self.padding
            else:
                break
            item_num += 1

        # While we're here, might as well calculate the highest top.
        self._the_highest_top = len(self.items) - 1
        item_num = self._the_highest_top
        y = mydest.bottom
        while y >= mydest.top and item_num >= 0:
            y -= self.items[item_num].height
            if y >= mydest.top:
                self._the_highest_top = item_num
            item_num -= 1
            y -= self.padding

    def render(self, scr, do_extras=True):
        mydest = self.get_rect()
        if do_extras:
            if self.predraw:
                self.predraw(scr)

            if self.border:
                self.border.render(mydest)

        scr.set_clip(mydest)
        self.arrange()
        for item_num, area in list(self._item_rects.items()):
            self.items[item_num].render(scr, area, (item_num == self.selected_item) and do_extras)

        scr.set_clip(None)

        if self.descobj:
            self.descobj(self.get_current_item())

    def get_mouseover_item(self, pos):
        # Return the menu item under this mouse position.
        # self.arrange must have been called previously!
        for item_num, area in list(self._item_rects.items()):
            if area.collidepoint(pos):
                return item_num

    def query(self):
        # TODO refactor this, so it doesnt hook the game loop anymore

        # A return of False means selection was cancelled.
        # pygame.event.clear()
        if not self.items:
            return False
        elif self.selected_item >= len(self.items):
            self.selected_item = 0
        self.no_choice_made = True
        self.choice = False

        # Do an initial arrangement of the menu.
        self.arrange()

    def proc_event(self, pc_input, source):
        """
        since may 2022,
        the new way to return "choice" values, is via posting an event
        """
        # while no_choice_made:
        #     pc_input = wait_event()

        # if pc_input.type == TIMEREVENT:
        #     # Redraw the menu on each timer event.
        #     self.render()
        #     kengi.flip()

        if pc_input.type == pygame.KEYDOWN:
            # A key was pressed, oh happy day! See what key it was and act
            # accordingly.
            if pc_input.key == pygame.K_UP:
                self.selected_item -= 1
                if self.selected_item < 0:
                    self.selected_item = len(self.items) - 1
                if self.selected_item not in self._item_rects:
                    self.top_item = min(self.selected_item, self._the_highest_top)
            elif pc_input.key == pygame.K_DOWN:
                self.selected_item += 1
                if self.selected_item >= len(self.items):
                    self.selected_item = 0
                if self.selected_item not in self._item_rects:
                    self.top_item = min(self.selected_item, self._the_highest_top)
            elif pc_input.key == pygame.K_SPACE or pc_input.key == pygame.K_RETURN:
                self.pev(event.EngineEvTypes.CONVCHOICE, value=self.items[self.selected_item].value)
                self.no_choice_made = False
            elif (pc_input.key == pygame.K_ESCAPE or pc_input.key == pygame.K_BACKSPACE) and self.can_cancel:
                self.no_choice_made = False
            elif 0 <= pc_input.key < 256 and chr(pc_input.key) in self.quick_keys:
                self.pev(event.EngineEvTypes.CONVCHOICE, value=self.quick_keys[chr(pc_input.key)])
                self.no_choice_made = False
            elif pc_input.key > 255 and pc_input.key in self.quick_keys:
                self.pev(event.EngineEvTypes.CONVCHOICE, value=self.quick_keys[pc_input.key])
                self.no_choice_made = False

        elif pc_input.type == pygame.MOUSEBUTTONDOWN:
            if pc_input.button == 1:
                mouse_pos = core.proj_to_vscreen(pygame.mouse.get_pos())

                moi = self.get_mouseover_item(mouse_pos)
                if moi is not None:

                    self.set_item_by_position(moi)
            elif pc_input.button == 4:
                self.top_item = max(self.top_item - 1, 0)
            elif pc_input.button == 5:
                self.top_item = min(self.top_item + 1, self._the_highest_top)

        elif pc_input.type == pygame.MOUSEBUTTONUP:
            mouse_pos = core.proj_to_vscreen(pygame.mouse.get_pos())
            if pc_input.button == 1:
                moi = self.get_mouseover_item(mouse_pos)
                if moi is self.selected_item:

                    self.pev(EngineEvTypes.CONVCHOICE, value=self.items[self.selected_item].value)
                    self.no_choice_made = False
            elif pc_input.button == 3 and self.can_cancel:
                self.no_choice_made = False

        elif pc_input.type == pygame.MOUSEMOTION:
            mouse_pos = core.proj_to_vscreen(pygame.mouse.get_pos())
            moi = self.get_mouseover_item(mouse_pos)
            if moi is not None:
                self.set_item_by_position(moi)

        elif pc_input.type == pygame.QUIT:
            self.no_choice_made = False

    def sort(self):
        self.items.sort()

    alpha_key_sequence = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"

    def add_alpha_keys(self):
        # Adds a quick key for every item currently in the menu.
        key_num = 0
        for item in self.items:
            item.msg = self.alpha_key_sequence[key_num] + ') ' + item.msg
            self.quick_keys[self.alpha_key_sequence[key_num]] = item.value
            key_num += 1
            if key_num >= len(self.alpha_key_sequence):
                break

    def add_files(self, filepat):
        file_list = glob.glob(filepat)

        for f in file_list:
            self.add_item(f, f)
        self.sort()

    def reposition(self):
        self.arrange()
        if self.selected_item < self.top_item:
            self.top_item = self.selected_item
        elif self.selected_item > max(self._item_rects.keys()):
            self.top_item = max(list(self._item_rects.keys()) + [self._the_highest_top])
        self.arrange()

    def has_value(self, v):
        for i in self.items:
            if i.value == v:
                return True

    def set_item_by_value(self, v):
        for n, i in enumerate(self.items):
            if i.value == v:
                self.selected_item = n
        self.reposition()

    def set_item_by_position(self, n):
        if n < len(self.items):
            self.selected_item = n
        self.reposition()

    def set_random_item(self):
        if self.items:
            n = random.randint(0, len(self.items) - 1)
            self.set_item_by_position(n)

    def get_current_item(self):
        if self.selected_item < len(self.items):
            return self.items[self.selected_item]

    def get_current_value(self):
        if self.selected_item < len(self.items):
            return self.items[self.selected_item].value


class PopUpMenu(Menu):
    """Creates a small menu at the current mouse position."""

    def __init__(self, w=200, h=250, predraw=None, border=default_border, **kwargs):
        mouse_pos = core.proj_to_vscreen(pygame.mouse.get_pos())
        x, y = mouse_pos
        x += 8
        y += 8
        sw, sh = core.get_screen().get_size()
        if x + w + 32 > sw:
            x += -w - 32
        if y + h + 32 > sh:
            y += -h - 32

        super().__init__(x, y, w, h, ANCHOR_UPPERLEFT, border=border, predraw=predraw, **kwargs)


class AlertMenu(Menu):
    WIDTH = 350
    HEIGHT = 250
    MENU_HEIGHT = 75

    FULL_RECT = Frect(-175, -125, 350, 250)
    TEXT_RECT = Frect(-175, -125, 350, 165)

    def __init__(self, desc):
        super().__init__(-self.WIDTH // 2, self.HEIGHT // 2 - self.MENU_HEIGHT, self.WIDTH, self.MENU_HEIGHT,
                         border=None, predraw=self.pre)
        self.desc = desc

    def pre(self):
        default_border.render(self.FULL_RECT.get_rect())
        draw_text(self.font, self.desc, self.TEXT_RECT.get_rect(), justify=0)



import json

from . import rpgmenu
from ... import _hub as kengi
from ... import event
from ... import pal

# - aliases
frects = kengi.polarbear.frects
kengi.polarbear.draw_text
draw_text = kengi.polarbear.draw_text
pygame = kengi.pygame

EngineEvTypes = event.EngineEvTypes
EventReceiver = event.EventReceiver


class Offer(object):
    # An Offer is a single line spoken by the NPC.
    # "effect" is a callable with no parameters.
    # "replies" is a list of replies.
    def __init__(self, msg, effect=None, replies=()):
        self.msg = msg
        self.effect = effect
        self.replies = list(replies)

    def __str__(self):
        return self.msg

    @classmethod
    def from_json(cls, jdict):
        # We spoke about not needing a json loader yet. But, in terms of hardcoding a conversation, it was just as
        # easy to write this as to hand-code a dialogue tree.
        msg = jdict.get("msg", "Hello there!")
        effect = None
        replies = list()
        for rdict in jdict.get("replies", ()):
            replies.append(Reply.from_json(rdict))
        return cls(msg, effect, replies)

    @classmethod
    def load_jsondata(cls, jsondata):
        return cls.from_json(json.loads(jsondata))


class Reply(object):
    # A Reply is a single line spoken by the PC, leading to a new offer
    def __init__(self, msg, destination=None):
        self.msg = msg
        self.destination = destination

    def __str__(self):
        return self.msg

    def apply_to_menu(self, mymenu):
        mymenu.add_item(self.msg, self.destination)

    @classmethod
    def from_json(cls, jdict):
        msg = jdict.get("msg", "And you too!")
        destination = jdict.get("destination")
        if destination:
            destination = Offer.from_json(destination)
        return cls(msg, destination)


class ConversationView(EventReceiver):
    # The visualizer is a class used by the conversation when conversing.
    # It has a "text" property and "render", "get_menu" methods.
    TEXT_AREA = frects.Frect(-75, -100, 300, 100)
    MENU_AREA = frects.Frect(-75, 30, 300, 80)
    PORTRAIT_AREA = frects.Frect(-240, -110, 150, 225)

    BGCOL = pal.punk['nightblue'] # '#221031'

    DEBUG = True

    @property
    def primitive_style(self):
        return self._primitive_style

    @primitive_style.setter
    def primitive_style(self, v):
        self._primitive_style = v

        # modify locations as we know that (Primitive style => upscaling is set to x3)
        x = 48 #58
        w = 192
        self.refxxx = x

        self.TEXT_AREA = frects.Frect(-x, -66, w, 80)
        self.MENU_AREA = frects.Frect(-x, 33, w, 80)
        self.PORTRAIT_AREA = frects.Frect(-x-100, -66, 90, 128)

        # optim:
        self.dim_menux = w+104

        self.text_rect = self.TEXT_AREA.get_rect()
        self.menu_rect = self.MENU_AREA.get_rect()
        self.portrait_rect = self.PORTRAIT_AREA.get_rect()
        self.taquet_portrait = self.portrait_rect[0]
        self.glob_rect = pygame.Rect(self.portrait_rect[0], 53, self.dim_menux, 182)


    def _primitiv_render(self, refscreen):
        pygame.draw.rect(refscreen, self.BGCOL, self.glob_rect)  # fond de fenetre

        pygame.draw.rect(refscreen, self.BGCOL, self.text_rect)
        draw_text(self.font, self.text, self.text_rect)

        if self.portrait:
            refscreen.blit(self.portrait, self.portrait_rect)

        pygame.draw.rect(refscreen, self.BGCOL, self.menu_rect)
        if self.existing_menu:
            self.existing_menu.render(refscreen)
        # pourtour menu
        pygame.draw.rect(refscreen, pal.punk['darkpurple'], (self.taquet_portrait, 148, self.dim_menux, 88), 2)

    def __init__(self, root_offer, chosen_font, ft_size, portrait_fn=None, pre_render=None):
        super().__init__()
        # - slight optim:
        self.text_rect = None
        self.menu_rect = None
        self.portrait_rect = None

        self._primitive_style = False  # can be used to make things faster for webctx
        self.zombie = False

        self.text = ''
        self.root_offer = root_offer
        self.pre_render = pre_render
        self.font = pygame.font.Font(chosen_font, ft_size)
        if portrait_fn:
            self.portrait = pygame.image.load(portrait_fn).convert_alpha()
        else:
            self.portrait = None

        self.curr_offer = root_offer
        self.dialog_upto_date = False
        self.existing_menu = None

    # def turn_off(self):  # because the conversation view can be closed from "outside" i.e. the main program
    #     if self.DEBUG:
    #         print('Conversation view closed')
    #
    #     if self.existing_menu:
    #         self.existing_menu.turn_off()
    #     super().turn_off()

    def update_dialog(self):
        if self.zombie:
            return

        if self.curr_offer is None:
            # auto-close everything
            self.pev(EngineEvTypes.CONVENDS)
            self.turn_off()
            self.zombie = True
            return

        if self.dialog_upto_date:
            return
        self.dialog_upto_date = True
        self.text = self.curr_offer.msg

        # create a new Menu inst.
        self.existing_menu = rpgmenu.Menu(
            self.MENU_AREA.dx, self.MENU_AREA.dy, self.MENU_AREA.w, self.MENU_AREA.h,
            border=None, predraw=None, font=self.font)  # predraw: self.render

        mymenu = self.existing_menu
        for i in self.curr_offer.replies:
            i.apply_to_menu(mymenu)
        if self.text and not mymenu.items:
            mymenu.add_item("[Continue]", None)
        else:
            mymenu.sort()

        nextfx = self.curr_offer.effect
        if nextfx:
            nextfx()

    def proc_event(self, ev, source):
        if self.existing_menu:
            self.existing_menu.proc_event(ev, None)  # forward event to menu if it exists

        if ev.type == EngineEvTypes.PAINT:
            self.render(ev.screen)

        elif ev.type == EngineEvTypes.LOGICUPDATE:
            self.update_dialog()

        elif ev.type == EngineEvTypes.CONVCHOICE:  # ~ iterate over the conversation...
            print('ConvChoice event received by', self.__class__.__name__)
            self.dialog_upto_date = False
            self.curr_offer = ev.value

    def _reg_render(self, refscreen):
        if self.pre_render:
            self.pre_render()
        text_rect = self.TEXT_AREA.get_rect()
        dborder = kengi.polarbear.default_border

        dborder.render(text_rect)
        draw_text(self.font, self.text, text_rect)
        dborder.render(self.MENU_AREA.get_rect())

        if self.existing_menu:
            self.existing_menu.render(refscreen)

        if self.portrait:
            refscreen.blit(self.portrait, self.PORTRAIT_AREA.get_rect())

    def render(self, scr):
        if self.primitive_style:
            self._primitiv_render(scr)
        else:
            self._reg_render(scr)



from ... import _hub
from ...compo import vscreen as core


pygame = _hub.pygame
Tilesets = _hub.tmx.data.Tilesets


class IsometricMapCursor:
    """
    I haven't updated this for the new map system yet... I will do that ASAP, even though the QuarterCursor is
    the one that will be useful for Niobepolis.
    """

    def __init__(self, x, y, image, frame=0, visible=True):
        self.x = x
        self.y = y
        self.image = image
        self.frame = frame
        self.visible = visible

    def render(self, dest: pygame.Rect):
        if self.visible:
            self.image.render(dest, self.frame)

    def set_position(self, scene, x, y, must_be_visible=True):
        if scene.on_the_map(x, y) and (scene.get_visible(x, y) or not must_be_visible):
            self.x, self.y = x, y

    def update(self, view, ev):
        if ev.type == pygame.MOUSEMOTION:
            self.set_position(view.isometric_map, *view._mouse_tile)
        elif ev.type == pygame.KEYDOWN:
            if ev.key == pygame.K_KP8:
                self.set_position(view.isometric_map, self.x - 1, self.y - 1)
                view.focus(self.x, self.y)
            elif ev.key == pygame.K_KP9:
                self.set_position(view.isometric_map, self.x, self.y - 1)
                view.focus(self.x, self.y)
            elif ev.key == pygame.K_KP6:
                self.set_position(view.isometric_map, self.x + 1, self.y - 1)
                view.focus(self.x, self.y)
            elif ev.key == pygame.K_KP3:
                self.set_position(view.isometric_map, self.x + 1, self.y)
                view.focus(self.x, self.y)
            elif ev.key == pygame.K_KP2:
                self.set_position(view.isometric_map, self.x + 1, self.y + 1)
                view.focus(self.x, self.y)
            elif ev.key == pygame.K_KP1:
                self.set_position(view.isometric_map, self.x, self.y + 1)
                view.focus(self.x, self.y)
            elif ev.key == pygame.K_KP4:
                self.set_position(view.isometric_map, self.x - 1, self.y + 1)
                view.focus(self.x, self.y)
            elif ev.key == pygame.K_KP7:
                self.set_position(view.isometric_map, self.x - 1, self.y)
                view.focus(self.x, self.y)


class IsometricMapQuarterCursor:
    # A cursor that only takes up one quarter of a tile.
    def __init__(self, x, y, surf, layer, visible=True):
        self._doublex = int(x * 2)
        self._doubley = int(y * 2)
        self.surf = surf
        self.layer_name = layer.name
        self.visible = visible

    def render(self, view):
        if self.visible:
            sx, sy = view.screen_coords(*self.get_pos())
            mylayer = view.isometric_map.get_layer_by_name(self.layer_name)
            mydest = self.surf.get_rect(midtop=(sx + mylayer.offsetx, sy + mylayer.offsety - 2))
            view.screen.blit(self.surf, mydest)

    def set_position(self, view, x, y):
        self._doublex = int(x * 2)
        self._doubley = int(y * 2)

    @property
    def x(self):
        return self._doublex // 2

    @property
    def y(self):
        return self._doubley // 2

    def get_pos(self):
        return float(self._doublex) / 2.0, float(self._doubley) / 2.0

    def focus(self, view):
        view.focus(*self.get_pos())

    def update(self, view, ev):
        if ev.type == pygame.MOUSEMOTION:
            mouse_x, mouse_y = core.proj_to_vscreen(pygame.mouse.get_pos())
            self.set_position(view, view.map_x(mouse_x, mouse_y, return_int=False),
                              view.map_y(mouse_x, mouse_y, return_int=False))
        elif ev.type == pygame.KEYDOWN:
            if ev.key == pygame.K_KP8:
                self._doublex -= 1
                self._doubley -= 1
                self.focus(view)
            elif ev.key == pygame.K_KP9:
                self._doubley -= 1
                self.focus(view)
            elif ev.key == pygame.K_KP6:
                self._doublex += 1
                self._doubley -= 1
                self.focus(view)
            elif ev.key == pygame.K_KP3:
                self._doublex += 1
                self.focus(view)
            elif ev.key == pygame.K_KP2:
                self._doublex += 1
                self._doubley += 1
                self.focus(view)
            elif ev.key == pygame.K_KP1:
                self._doubley += 1
                self.focus(view)
            elif ev.key == pygame.K_KP4:
                self._doublex -= 1
                self._doubley += 1
                self.focus(view)
            elif ev.key == pygame.K_KP7:
                self._doublex -= 1
                self.focus(view)



import re
import textwrap

from .. import _hub as inj
from ..compo import vscreen as core


pygame = inj.pygame

# - options
# --------------- constants for IgCustomConsole display formattage -------------
CONSOLE_FT_SIZE = 12
omega_ascii_letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
ANTIALIAS_OPT = False


# - addons - - > DU TAF PROSPECTIF
# EventReicever = kengi.event.EventReceiver
#
# class IgConsoleCtrl(EventReicever):
#     def proc_event(self, ev, source):
#         pass
#
# class IgConsoleView(EventReicever):
#     def proc_event(self, ev, source):
#         pass


# --------------- SYNTAX DETECTORS FOR IgCustomConsole --------------
re_token = re.compile(r"""[\"].*?[\"]|[\{].*?[\}]|[\(].*?[\)]|[\[].*?[\]]|\S+""")
re_is_list = re.compile(r'^[{\[(]')

# re_is_number = re.compile(r"""
# (?x)
# [-]?[0][x][0-9a-fA-F]+[lLjJ]? | 	#  Hexadecimal
# [-]?[0][0-7]+[lLjJ]? |				#  Octal
# [-]?[\d]+(?:[.][\d]*)?[lLjJ]?		#  Decimal (Int or float)
#""")
re_is_number = re.compile("(?x)[-]?[\d]+(?:[.][\d]*)?[lLjJ]?")  # decimal only

re_is_assign = re.compile(r'[$](?P<name>[a-zA-Z_]+\S*)\s*[=]\s*(?P<value>.+)')
re_is_comment = re.compile(r'\s*#.*')
re_is_var = re.compile(r'^[$][a-zA-Z_]+\w*\Z')


# -------------------- outils pr implementation CustomIgConsole -------------------
class ParseError(Exception):
    def __init__(self, token):
        self.token = token

    def at_token(self):
        return self.token


def balanced(t):  # UTIL func used in .tokenize
    stack = []
    pairs = {"\'": "\'", '\"': '\"', "{": "}", "[": "]", "(": ")"}
    for char in t:
        if stack and char == pairs[stack[-1]]:
            stack.pop()
        elif char in pairs:
            stack.append(char)
    return not bool(stack)


# ----------------------- classe essentielle au plugin -----------------------------
class CustomConsole:
    """
    en phase de devenir le MODELE, a separer de la vue...
    """

    def __init__(self, screen, rect, functions=None, key_calls=None, vari=None, syntax=None, fontobj=None):
        """
        :param screen:
        :param rect:
        :param functions:
        :param key_calls:
        :param vari:
        :param syntax: dict that associates v,k where
           v is a regexp processor
           k is a function in the form console_func(console, match) that calls console.output(...)
        """
        self.message_of_the_day = ["no Motd"]
        self.scrref = screen
        self.li_raw_lines_labels = list()

        self.bg_color = '#06170e'  # very dark green

        self.bg_alpha = 200  # how transparent is the Console Background, 255 -> opaque
        self.txt_color_i = (0, 255, 0)  # Green bright Input Color
        self.txt_color_o = (0, 153, 0)  # Dark Green output Color

        self.changed = True
        self.active = False
        self.preserve_events = True
        self.repeat_rate = [500, 30]

        self.ps1 = "x "
        self.ps2 = ">>> "  # invite cmd
        self.ps3 = "... "
        self.pos_sym_char = "_"
        self.line_disp_yoffset = -2  # how many pixels shll i move it?

        self.c_ps = self.ps2

        # --- c_out is the history of all text
        # -- c_hist is the history of all commands
        self.c_out = ['',]
        self.c_in = ""
        self.c_hist = [""]
        self.c_hist_pos = 0
        self.c_pos = 0
        self.c_draw_pos = 0
        self.c_scroll = 0

        #self.parent_screen = screen

        self.rect = pygame.Rect(rect)
        self.size = self.rect.size

        if fontobj is None:
            self.font = pygame.font.Font(None, 18)
        else:
            self.font = fontobj
        # en remplacement de:
        # adhoc_ft_size = CONSOLE_FT_SIZE if (ftsize is None) else ftsize
        # print('CONSOLE uses font {}, size{}'.format(fontpath, adhoc_ft_size))
        # self.font = pygame.font.Font(fontpath, adhoc_ft_size)

        self.font_height = self.font.get_linesize()
        self.max_lines = int((self.size[1] / self.font_height) - 1)
        ftdim = self.font.size(omega_ascii_letters)
        kappa = ftdim[0] / len(omega_ascii_letters)

        self.max_chars = int(((self.size[0]) / kappa) - 1)
        self.txt_wrapper = textwrap.TextWrapper()
        self.bg_layer = pygame.Surface(self.size)

        # TODO use alpha!
        # self.bg_layer.set_alpha(self.bg_alpha)

        self.txt_layer = pygame.Surface(self.size)
        self.txt_layer.set_colorkey(self.bg_color)

        pygame.key.set_repeat(*self.repeat_rate)

        self.key_calls = {}
        self.func_calls = {}
        self.user_vars = vari
        self.user_syntax = syntax
        self.user_namespace = {}

        self.add_key_calls({"l": self.clear, "c": self.clear_input, "w": self.toggle_active})
        self.add_key_calls(key_calls)

        self.add_functions_calls({"help": self.help, "echo": self.output, "clear": self.clear})
        self.add_functions_calls(functions)

        self.cb_func = None

    def set_motd(self, msg):
        self.message_of_the_day = msg.splitlines()
        self.c_out.extend(self.message_of_the_day)

    def add_functions_calls(self, functions):
        if isinstance(functions, dict):
            self.func_calls.update(functions)

    def add_key_calls(self, functions):
        if isinstance(functions, dict):
            self.key_calls.update(functions)

    def output(self, text):
        """Print a string on the Console. Use: echo "Test Test Test" """
        self.changed = True
        if not isinstance(text, str):
            text = str(text)
        text = text.expandtabs()
        text = text.splitlines()
        self.txt_wrapper.width = self.max_chars
        for line in text:
            for w in self.txt_wrapper.wrap(line):
                self.c_out.append(w)

    def submit_input(self, text):
        self.clear_input()
        if self.cb_func:
            self.output(text)
        else:
            self.output(self.c_ps + text)

        self.c_scroll = 0
        if self.cb_func:
            self.cb_func(text)
            self.cb_func = None
        else:
            self.send_pyconsole(text)

    def format_input_line(self):
        text = self.c_in[:self.c_pos] + self.pos_sym_char + self.c_in[self.c_pos + 1:]
        n_max = int(self.max_chars - len(self.c_ps))
        vis_range = self.c_draw_pos, self.c_draw_pos + n_max
        if self.cb_func:
            prefix = ''
        else:
            prefix = self.c_ps
        return prefix + text[vis_range[0]:vis_range[1]]

    def str_insert(self, text, strn):
        string = text[:self.c_pos] + strn + text[self.c_pos:]
        self.set_pos(self.c_pos + len(strn))
        return string

    def clear_input(self):
        self.c_in = ""
        self.c_pos = 0
        self.c_draw_pos = 0

    def set_pos(self, newpos):
        self.c_pos = newpos
        if (self.c_pos - self.c_draw_pos) >= int(self.max_chars - len(self.c_ps)):
            self.c_draw_pos = max(0, int(self.c_pos - (self.max_chars - len(self.c_ps))))
        elif self.c_draw_pos > self.c_pos:
            self.c_draw_pos = self.c_pos - (self.max_chars/2)
            if self.c_draw_pos < 0:
                self.c_draw_pos = 0
                self.c_pos = 0

    def toggle_active(self):
        self.active = not self.active

    def activate(self):
        self.active = True

    def desactivate(self):
        self.active = False

    def add_to_history(self, text):
        self.c_hist.insert(-1, text)
        self.c_hist_pos = len(self.c_hist) - 1

    def draw(self):
        if not self.active:
            return

        if self.changed:  # update text layer
            del self.li_raw_lines_labels[:]
            # creation du txt layer
            self.txt_layer.fill(self.bg_color)
            lines = self.c_out[-(self.max_lines + self.c_scroll):len(self.c_out) - self.c_scroll]
            y_pos = self.size[1]-(self.font_height*(len(lines)+1))
            for line in lines:
                label_surf = self.font.render(line, False, self.txt_color_o, (0, 0, 0))
                dpos = (1, y_pos+self.line_disp_yoffset)
                # self.txt_layer.blit(label_surf, dpos )
                self.li_raw_lines_labels.append([label_surf, dpos])
                y_pos += self.font_height

            last_label = self.font.render(self.format_input_line(), False, self.txt_color_i)
            last_pos = (1, self.size[1] - self.font_height + self.line_disp_yoffset)
            # self.txt_layer.blit(last_label, last_pos)
            self.li_raw_lines_labels.append([last_label, last_pos])

            # refresh bg_layer qui contiendra aussi txt_layer...
            # self.bg_layer = pygame.Surface(self.size)
            # self.bg_layer.blit(self.txt_layer, (0, 0)) #, 0, 0))
            self.changed = False

        self.bg_layer.fill(self.bg_color)
        self.scrref.blit(self.bg_layer, (0, 0))
        # self.scrref.blit(self.txt_layer, self.rect)
        for label, lpos in self.li_raw_lines_labels:
            self.scrref.blit(label, lpos)

    def process_input(self, eventlist):
        if not self.active:
            return

        for event in eventlist:
            if event.type == pygame.KEYDOWN:
                self.changed = True
                # Special Character Manipulation
                if event.key == pygame.K_TAB:
                    self.c_in = self.str_insert(self.c_in, "    ")
                elif event.key == pygame.K_BACKSPACE:
                    if self.c_pos > 0:
                        self.c_in = self.c_in[:self.c_pos - 1] + self.c_in[self.c_pos:]
                        self.set_pos(self.c_pos - 1)
                elif event.key == pygame.K_DELETE:
                    if self.c_pos < len(self.c_in):
                        self.c_in = self.c_in[:self.c_pos] + self.c_in[self.c_pos + 1:]
                elif event.key == pygame.K_RETURN or event.key == 271:
                    self.submit_input(self.c_in)
                # Changing Cursor Position
                elif event.key == pygame.K_LEFT:
                    if self.c_pos > 0:
                        self.set_pos(self.c_pos - 1)
                elif event.key == pygame.K_RIGHT:
                    if self.c_pos < len(self.c_in):
                        self.set_pos(self.c_pos + 1)
                elif event.key == pygame.K_HOME:
                    self.set_pos(0)
                elif event.key == pygame.K_END:
                    self.set_pos(len(self.c_in))
                # History Navigation
                elif event.key == pygame.K_UP:
                    if len(self.c_out):
                        if self.c_hist_pos > 0:
                            self.c_hist_pos -= 1
                        self.c_in = self.c_hist[self.c_hist_pos]
                        self.set_pos(len(self.c_in))
                elif event.key == pygame.K_DOWN:
                    if len(self.c_out):
                        if self.c_hist_pos < len(self.c_hist) - 1:
                            self.c_hist_pos += 1
                        self.c_in = self.c_hist[self.c_hist_pos]
                        self.set_pos(len(self.c_in))
                # Scrolling
                elif event.key == pygame.K_PAGEUP:
                    if self.c_scroll < len(self.c_out) - 1:
                        self.c_scroll += 1
                elif event.key == pygame.K_PAGEDOWN:
                    if self.c_scroll > 0:
                        self.c_scroll -= 1
                # Normal character printing
                elif event.key >= 32:
                    mods = pygame.key.get_mods()
                    if mods & pygame.KMOD_CTRL:
                        if event.key in range(256) and chr(event.key) in self.key_calls:
                            self.key_calls[chr(event.key)]()
                    else:
                        char = str(event.unicode)
                        self.c_in = self.str_insert(self.c_in, char)

    def send_pyconsole(self, text):
        if not text:  # Output a blank row if nothing is entered
            self.output("")
            return

        self.add_to_history(text)

        # Determine if the statement is an assignment
        assign = re_is_assign.match(text)

        # If it is tokenize only the "value" part of $name = value
        if assign:
            tokens = self.tokenize(assign.group('value'))
        else:
            tokens = self.tokenize(text)

        if tokens is None:
            return

        # Evaluate
        try:
            out = None
            # A variable alone on a line
            if (len(tokens) == 1) and re_is_var.match(text) and not assign:
                out = tokens[0]
            # Statement in the form $name = value
            elif (len(tokens) == 1) and assign:
                self.setvar(assign.group('name'), tokens[0])
            else:
                # Function
                if tokens[0] not in self.func_calls:
                    self.output("Unknown Command: " + str(tokens[0]))
                    self.output(r'Type "help" for a list of commands.')
                else:
                    out = self.func_calls[tokens[0]](*tokens[1:])
                    if out is None:
                        print('*Warning console func MUST return a value*')
                    # Assignment from function's return value
                    if assign:
                        self.setvar(assign.group('name'), out)
                    if out is not None:
                        self.output(out)

        except ValueError as ve:
            self.output("Error: {0}".format(ve))
        except TypeError as te:
            tmp = "Error {0}".format(te)
            if tmp.find('missing'):
                self.output("Bad syntax, type help cmd for more info")
            else:
                self.output(tmp)

    def setvar(self, name, value):
        """Sets the value of a variable"""
        if name in self.user_vars or name not in self.__dict__:
            self.user_vars.update({name: value})
            self.user_namespace.update(self.user_vars)
        elif name in self.__dict__:
            self.__dict__.update({name: value})

    def convert_token(self, tok):
        """
        TODO tomanalyz
        -this is one of the most important func
        :param tok:
        :return:
        """
        tok = tok.strip("$")
        try:
            tmp = eval(tok, self.__dict__, self.user_namespace)
            return tmp

        except SyntaxError as strerror:
            self.output("SyntaxError: " + str(strerror))
            raise ParseError(tok)

        except TypeError as strerror:
            self.output("TypeError: " + str(strerror))
            raise ParseError(tok)

        except NameError as strerror:
            self.output("NameError: " + str(strerror))

        self.output("Error:")
        raise ParseError(tok)

    def tokenize(self, s):
        if re_is_comment.match(s):
            return [s]

        for r in self.user_syntax:
            group = r.match(s)
            if group:
                self.user_syntax[r](self, group)
                return

        tokens = re_token.findall(s)
        tokens = [i.strip("\"") for i in tokens]
        cmd = []
        i = 0
        while i < len(tokens):
            t_count = 0
            val = tokens[i]

            if re_is_number.match(val):
                cmd.append(self.convert_token(val))
            elif re_is_var.match(val):
                cmd.append(self.convert_token(val))
            elif val == "True":
                cmd.append(True)
            elif val == "False":
                cmd.append(False)
            elif re_is_list.match(val):
                while not balanced(val) and (i + t_count) < len(tokens) - 1:
                    t_count += 1
                    val += tokens[i + t_count]
                else:
                    if (i + t_count) < len(tokens):
                        cmd.append(self.convert_token(val))
                    else:
                        raise ParseError(val)
            else:
                cmd.append(val)
            i += t_count + 1
        return cmd

    def clear(self):
        """Clear the screen! Use: clear"""
        del self.li_raw_lines_labels[:]
        self.c_out = [" "]
        self.c_scroll = 0
        return "[Screen Cleared]"

    def help(self, *args):
        if args:
            items = list()
            for cfunc_name in self.func_calls:
                if cfunc_name in args:
                    e = (cfunc_name, self.func_calls[cfunc_name])
                    items.append(e)

            for name, v in items:
                nb_args = v.__code__.co_argcount
                if nb_args > 0:
                    nb_args -= 1 if v.__code__.co_varnames[0] == "self" else 0
                out = "{} :: Takes {} arguments,".format(name, nb_args)
                doc = v.__doc__
                if doc:
                    out += textwrap.dedent(doc)
                tmp_indent = self.txt_wrapper.subsequent_indent
                self.txt_wrapper.subsequent_indent = " " * (len(name) + 2)
                self.output(out)
                self.txt_wrapper.subsequent_indent = tmp_indent
        else:
            out = "Available commands are\n"
            lc = list(self.func_calls.keys())
            lc.sort()
            out += '; '.join(lc)
            self.output(out)
            self.output(r'Type "help command-name" for more information on that command')



import katagames_sdk.api as katapi
import katagames_sdk.capsule.engine_ground.conf_eng as cgmconf
from katagames_sdk.capsule.pygame_provider import get_module
from katagames_sdk.engine import EventReceiver, EngineEvTypes


pygame = get_module()


class KataFrameC(EventReceiver):

    def __init__(self, ref_mod, ref_view):
        super().__init__()
        self.ref_mod = ref_mod
        self.ref_view = ref_view
        self.glvars_module = None

    def proc_event(self, ev, source):

        if ev.type == EngineEvTypes.DOAUTH:
            self._handle_auth()

        elif ev.type == pygame.KEYDOWN:
            self._handle_keypress(ev)

        elif ev.type == pygame.QUIT:
            self.pev(EngineEvTypes.POPSTATE)

    def _handle_auth(self):
        login_feedback = katapi.try_auth_server(self.ref_mod.username, self.ref_mod.plainpwd)

        if login_feedback:
            username, acc_id = self.ref_mod.username, int(login_feedback)
            self.glvars_module.username = username
            self.glvars_module.acc_id = acc_id

            if cgmconf.runs_in_web():
                print('commit to session_storage ->ye')
                katapi.save_to_local_session(username, acc_id)
            self.pev(EngineEvTypes.CHANGESTATE, state_ident=0)

        else:
            self.ref_mod.reset_fields()
            # gestion refus -> affiche un message erreur
            self.ref_view.sig_bad_auth()

    def _handle_keypress(self, ev):
        if ev.key == pygame.K_RETURN or ev.key == pygame.K_KP_ENTER:
            pass

        elif ev.key == pygame.K_ESCAPE:
            self.pev(EngineEvTypes.POPSTATE)

        elif ev.key == pygame.K_TAB:
            return  # sert à ignorer caractères TAB ds la saisie

        elif ev.key == pygame.K_BACKSPACE:
            self.ref_mod.suppr_lettre()

        else:
            self.ref_mod.input_char(ev.unicode)  # register the key



"""
Katasdk allows the creation of games that are super easy to share online

Copyright (C) 2022 Gaudia Tech Inc.

contac@kata.games
twitter.com/CreatePlayEarn
"""

SDKVER_TAG = '0.0.9'

from . import katagames_engine as _kengi

_kengi.get_injector().package_arg = 'katagames_sdk.katagames_engine'
kengi = _kengi



"""
Keep this file separate from __init__.py!
its important, bc all sub-modules in kengi may import _hub
in order to refer to other dependencies/sub-modules
"""
from .Injector import Injector


kengi_inj = Injector({
    'ai': '.looparts.ai',
    'anim': '.looparts.anim',
    'ascii': '.looparts.ascii',
    'demolib': '.looparts.demolib',
    'polarbear': '.looparts.polarbear',
    'console': '.looparts.console',
    'gui': '.looparts.gui',
    'isometric': '.looparts.isometric',
    'legacy': '.looparts.legacy',
    'tabletop': '.looparts.tabletop',

    # nota bene: pygame is dynamically added to this dict,
    # its done elsewhere
    'tmx': '.looparts.tmx',

    'rogue': '.looparts.rogue',
    'terrain': '.looparts.terrain',
})


def __getattr__(targ_sm_name):
    if targ_sm_name in kengi_inj:
        return kengi_inj[targ_sm_name]
    else:
        raise AttributeError(f"kengi has no attribute named {targ_sm_name}")



VYEAR_2DIG = 22  # remember to use only 2 digits
VMONTH = 8
_idx = 1  # 1 =alpha, 2 =beta, 3 =release candidate, 0 =legit release
VRANK = 2

"""
We target this standard format:
- Alpha release     yy.MMaN
- Beta release      yy.MMbN
- Release candidate yy.MMrcN
- Final release     yy.MM

So, to respect the PEP440 do not change anything pass this line!
"""
VRTYPE = ['', 'a', 'b', 'rc'][_idx]  # do not change this line
_suffix = '' if '' == VRTYPE else VRTYPE + str(VRANK)
ENGI_VERSION = f"{VYEAR_2DIG}.{VMONTH}{_suffix}"



"""
sub-module dedicated to GUI management/creation,
 belongs to the kata engine
"""

from .Button import Button
from .Button import ButtonPanel
from .Button2 import Button2
from .Label import Label
from .TextBlock import TextBlock
from .Trigger import Trigger
from .WidgetBo import WidgetBo
from .text import ImgBasedFont
from ... import _hub as inj
from ... import event


pygame = inj.pygame
# TODO doublon Label , Etiquette
#  à regler


# ---------- pr GUI
class Etiquette:
    ft_obj = None

    def __init__(self, text, pos, rgb_color):
        if self.ft_obj is None:
            raise ValueError('use set_font(...) first! ')

        self._text = text
        self.pos = pos
        self._color = rgb_color
        self.img = self.ft_obj.render(self._text, True, self._color)

    @classmethod
    def set_font(cls, ft_obj):
        cls.ft_obj = ft_obj

    def get_text(self):
        return self._text

    def set_text(self, t):
        self._text = t
        self.img = self.ft_obj.render(self._text, True, self._color)


class TextInput(event.EventReceiver):
    """
    Simple text entry component.
    """

    _caret_color = (90, 90, 90)
    _padding = 4  # en px

    def __init__(self, nickname, font, how_to_process_cb, pos, width=300):
        """
        nickname: nickname, all messages will be prefixed with the nickname
        font    : pygame.font.Font object
        width   : in pixel that this element can use (this restricts the number
                  of char you can enter

        events:
        in : pygame.KEYDOWN
        out: eventtypes.CHATMSG
        """
        super().__init__()

        self.position = pos
        # test que le 3e arg est callable...
        assert hasattr(how_to_process_cb, '__call__')
        self.on_enter_func = how_to_process_cb

        self.pwd_field = False

        # data
        self.__txt_content = ""
        self.caretpos = 0
        self.max = 255
        self.nickname = nickname

        # drawing
        self.dirty = True
        self.font = font
        height = self.font.get_ascent() - self.font.get_descent() + 8
        self.image = pygame.Surface((width, height)).convert()
        self.size = (width, height)

        self.text_color = (1, 1, 1)
        self.text_field_rect = pygame.Rect(0, 0, width - 1, height - 1)
        self.text_img = pygame.Surface((2, 2))
        self.pixel_width = width - 4

        self._focus = None
        self.fill_color = None
        self.no_focus()

    def has_focus(self):
        return self._focus

    def get_disp_text(self):
        if self.pwd_field:
            return TextInput.hide_text(self.__txt_content)
        return self.__txt_content

    def focus(self):
        self._focus = True
        self.fill_color = (220, 220, 220)
        self.render_field()

    def no_focus(self):
        self._focus = False
        self.fill_color = (100, 100, 100)
        self.render_field()

    def contains(self, scr_pos):
        w, h = self.image.get_size()
        a, b = self.position[0], self.position[0] + w
        c, d = self.position[1], self.position[1] + h
        x, y = scr_pos
        if (a < x < b) and (c < y < d):
            return True
        return False

    @staticmethod
    def hide_text(txt_content):
        tmp = ['*' for i in range(len(txt_content))]
        return ''.join(tmp)

    def proc_event(self, event, source):
        if event.type != pygame.KEYDOWN:
            return

        # - traitement touche pressée
        if event.key == pygame.K_RETURN:
            # self.on_enter()
            self.on_enter_func(self.__txt_content)
            self.__txt_content = ''
            self.caretpos = 0

        elif event.key == pygame.K_RIGHT:
            self.move_caret(+1)

        elif event.key == pygame.K_LEFT:
            self.move_caret(-1)

        elif event.key == pygame.K_HOME:
            self.move_caret('home')

        elif event.key == pygame.K_END:
            self.move_caret('end')

        elif event.key == pygame.K_BACKSPACE:
            self.backspace_char()

        elif event.key == pygame.K_DELETE:
            self.delete_char()

        elif event.key == pygame.K_TAB:
            pass

        else:
            if event.unicode != '':
                if len(self.__txt_content) < self.max:
                    self.__txt_content = self.__txt_content[:self.caretpos] + event.unicode + self.__txt_content[
                                                                                              self.caretpos:]
                    self.caretpos += 1
        self.render_field()

    def move_caret(self, steps):
        """
        Moves the caret about steps. Positive numbers moves it right, negative
        numbers left.
        """
        if steps == 'home':
            self.caretpos = 0
        elif steps == 'end':
            self.caretpos = len(self.__txt_content)
        else:
            assert isinstance(steps, int)
            self.caretpos += steps

        if self.caretpos < 0:
            self.caretpos = 0
        if self.caretpos > len(self.__txt_content):
            self.caretpos = len(self.__txt_content)

    def backspace_char(self):
        """
        Deltes the char befor the caret position.
        """
        if self.caretpos > 0:
            self.__txt_content = self.__txt_content[:self.caretpos - 1] + self.__txt_content[self.caretpos:]
            self.caretpos -= 1

    def delete_char(self):
        """
        Deltes the char after the caret position.
        """
        self.__txt_content = self.__txt_content[:self.caretpos] + self.__txt_content[self.caretpos + 1:]

    def render_field(self):
        """
        Renders the string to self.image.
        """
        self.image.fill(self.fill_color)
        content = self.get_disp_text()

        if len(content):
            # while self.font.size(content)[0] > self.pixel_width:
            #    self.backspace_char()
            self.text_img = self.font.render(content, 1, self.text_color, self.fill_color)
            self.image.blit(self.text_img, (2, 2))

            # - draw caret
            # TODO fix pygame draw line
            # xpos = self.font.size(content[:self.caretpos])[0] + 2
            # pygame.draw.line(self.image, self._caret_color, (xpos, self._padding),
            #                  (xpos, self.image.get_height() - self._padding), 2)

        pygame.draw.rect(self.image, (100, 100, 100), self.text_field_rect, 2)




"""
TODO support .json, uncompressed files
"""
from ... import _hub as inj
pygame = inj.pygame


class SpriteLayer(pygame.sprite.AbstractGroup):
    def __init__(self):
        super(SpriteLayer, self).__init__()
        self.visible = True

    def set_view(self, x, y, w, h, viewport_ox=0, viewport_oy=0):
        self.view_x, self.view_y = x, y
        self.view_w, self.view_h = w, h
        x -= viewport_ox
        y -= viewport_oy
        self.dx = viewport_ox
        self.dy = viewport_oy
        self.position = (x, y)

    def draw(self, screen):
        ox, oy = self.position
        w, h = self.view_w, self.view_h

        for sprite in self.sprites():
            sx, sy = sprite.rect.topleft
            # Only the sprite's defined width and height will be drawn
            area = pygame.Rect((0, 0),
                               (sprite.rect.width,
                                sprite.rect.height))
            screen.blit(sprite.image, (sx - ox, sy - oy), area)


class Viewport(pygame.Rect):
    """
    viewport: based upont pygame.Rect (pygame obj. for storing rectangular coordinates)

    can draw tilemaps, constrained by viewport origin and dimensions

    attributes:
    fx, fy - viewport focus point
    view_w, view_h - viewport size
    view_x, view_y - viewport offset (origin)
    """
    _old_focus = None

    def __init__(self,  tilemap_ref, origin, size, **kwargs):
        super().__init__(origin, size)
        self._tm = tilemap_ref  # Tilemap instance

        self.fx, self.fy = 0, 0  # viewport focus point
        self.view_w, self.view_h = size  # viewport size
        self.view_x, self.view_y = origin  # viewport offset

        # rn the followin line is very important as it is responible for
        # init. Layer draw process (forcing .set_view(...) call)
        # and to set offset properly!
        self.force_focus(*origin)
        # debug: print('force_focus: ', origin[0], origin[1])

        # this is like a temp storage
        self.childs_ox, self.childs_oy = None, None

        # wtf? how is this useful
        self.restricted_fx, self.restricted_fy = None, None

        for ckey, val in kwargs.items():
            setattr(self, ckey, val)

    def set_focus(self, fx, fy, force=False):
        """
        Determine the viewport based on a desired focus pixel in the
        Layer space (fx, fy) and honoring any bounding restrictions of
        child layers.

        The focus will always be shifted to ensure no child layers display
        out-of-bounds data, as defined by their dimensions px_width and px_height.
        """
        # -debug
        # print('set_focus CALL. Args= ', fx, fy)

        # The result is that all chilren will have their viewport set, defining
        # which of their pixels should be visible.
        fx, fy = int(fx), int(fy)
        self.fx, self.fy = fx, fy

        a = (fx, fy)

        # check for NOOP (same arg passed in)
        if not force and self._old_focus == a:
            # noop
            return
        self._old_focus = a

        # get our viewport information, scaled as appropriate
        w = int(self.view_w)
        h = int(self.view_h)
        w2, h2 = w // 2, h // 2

        totwidth, totheight = self._tm.pix_width, self._tm.pix_height
        if totwidth <= w:
            # this branch for centered view and no view jump when
            # crossing the center; both when world width <= view width
            restricted_fx = totwidth / 2
        else:
            if (fx - w2) < 0:
                restricted_fx = w2  # hit minimum X extent
            elif (fx + w2) > totwidth:
                restricted_fx = totwidth - w2  # hit maximum X extent
            else:
                restricted_fx = fx

        if self._tm.pix_height <= h:
            # this branch for centered view and no view jump when
            # crossing the center; both when world height <= view height
            restricted_fy = totheight / 2
        else:
            if (fy - h2) < 0:
                restricted_fy = h2  # hit minimum Y extent
            elif (fy + h2) > totheight:
                restricted_fy = totheight - h2  # hit maximum Y extent
            else:
                restricted_fy = fy

        # ... and this is our focus point, center of screen
        self.restricted_fx = int(restricted_fx)
        self.restricted_fy = int(restricted_fy)

        # determine child view bounds to match that focus point
        x, y = int(restricted_fx - w2), int(restricted_fy - h2)
        self._match_focus_pt(x, y, w, h)

    def force_focus(self, fx, fy):
        """
        Force the manager to focus on a point,
        regardless of any managed layer visible boundaries.
        """

        # This calculation takes into account the scaling of this Layer (and
        # therefore also its children).
        # The result is that all chilren will have their viewport set, defining
        # which of their pixels should be visible.
        self.fx, self.fy = list(map(int, (fx, fy)))

        # self.fx, self.fy = fx, fy

        # get our view size
        w = int(self.view_w)
        h = int(self.view_h)
        w2, h2 = w // 2, h // 2

        # bottom-left corner of the viewport
        x, y = fx - w2, fy - h2
        self._match_focus_pt(x, y, w, h)

    def _match_focus_pt(self, x, y, dimw, dimh):
        self.x = x  # self.x is a super(), that is Rect-defined attribute
        self.y = y  # self.y is a super(), that is Rect-defined attribute

        # change offsets, n update all layers view
        self.childs_ox = x - self.view_x
        self.childs_oy = y - self.view_y
        for layer in self._tm.layers:
            layer.set_view(x, y, dimw, dimh, self.view_x, self.view_y)

    def draw(self, screen):
        for layer in self._tm.layers:
            if layer.visible:
                layer.draw(screen)

    # - utils
    def pixel_from_screen(self, x, y):
        # look up the Layer-space pixel matching the screen-space pixel.
        vx, vy = self.childs_ox, self.childs_oy
        return int(vx + x), int(vy + y)

    def pixel_to_screen(self, x, y):
        # look up the screen-space pixel matching the Layer-space pixel
        screen_x = x - self.childs_ox
        screen_y = y - self.childs_oy
        return int(screen_x), int(screen_y)

    def index_at(self, x, y):
        # return the map index at the (screen-space) pixel position.
        sx, sy = self.pixel_from_screen(x, y)
        return int(sx // self._tm.tile_width), int(sy // self._tm.tile_height)
